# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
import biosteam as bst
from math import sqrt, log, exp
from biosteam.utils import MockStream
import thermosteam as tmo
from biorefineries.cane.composition import (
    set_lipid_fraction as set_oil_fraction,
    set_line_composition_parameters,
)
from biosteam import main_flowsheet, UnitGroup
from chaospy import distributions as shape
import numpy as np
from .feature_mockups import (
    all_parameter_mockups,
    all_metric_mockups, 
)
from .process_settings import load_process_settings
from .chemicals import create_cellulosic_oilcane_chemicals as create_chemicals
from biorefineries.cellulosic import PretreatmentReactorSystem as PRS
from .systems import (
    create_sugarcane_to_ethanol_system,
    create_sugarcane_to_sugar_and_ethanol_system,
    create_oilcane_to_biodiesel_and_ethanol_1g,
    create_oilcane_to_biodiesel_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation,
    create_sugarcane_to_ethanol_combined_1_and_2g,  
    create_oilcane_to_crude_oil_and_ethanol_1g,
    create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation,
    create_oilcane_to_biodiesel_1g,
    create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation,
    create_oilcane_to_biodiesel_and_actag_1g,
    create_oilcane_to_biodiesel_and_actag_combined_1_and_2g_post_fermentation_oil_separation,
)
from .data import microbial_oil_baseline as perf
from .parse_configuration import (
    parse_configuration,
    format_configuration,
)
from biorefineries.tea import (
    create_cellulosic_ethanol_tea as create_tea
)
from .composition import (
    CaneCompositionSpecification,
)
from .oil_extraction import (
    OilExtractionSpecification,
    MockExtractionSpecification,
)
from .data.lca_characterization_factors import (
    GWP_characterization_factors,
    set_GWPCF,
    GWP,
)

__all__ = (
    'Biorefinery',
    'YRCP2023',
)

PRS_cost_item = PRS.cost_items['Pretreatment reactor system']

energycane_dct = dict(
    Water=60,
    Sucrose=9.0,
    Glucose=0.1,
    Fiber=29.3,
    Ash=2.8,
)

kg_per_ton = 907.18474
kg_per_MT = 1000
L_per_gal = 3.7854
biodiesel_kg_per_gal = 3.3111
biodiesel_gal_per_kg = 1. / biodiesel_kg_per_gal
ethanol_kg_per_gal = 2.98668849
ethanol_gal_per_kg = 1. / ethanol_kg_per_gal
biodiesel_L_per_kg = biodiesel_gal_per_kg * L_per_gal
biodiesel_kg_per_L = 1. / biodiesel_L_per_kg
ethanol_L_per_kg = ethanol_gal_per_kg * L_per_gal
ethanol_kg_per_L = 1. / ethanol_L_per_kg

screwpress_microbial_oil_recovery = frozenset([9])
cellulosic_configurations = frozenset([-2, 2, 4, 6, 8, 9])
biodiesel_configurations = frozenset([1, 2, 5, 6, 7, 8, 9])
ethanol_configurations = frozenset([-2, -1, 1, 2, 3, 4])
actag_configurations = frozenset([10, 11])
conventional_ethanol_configurations = ethanol_configurations.difference(cellulosic_configurations)
cellulosic_ethanol_configurations = cellulosic_configurations.intersection(ethanol_configurations)
ethanol_biodiesel_configurations = ethanol_configurations.intersection(biodiesel_configurations)
cellulosic_ethanol_biodiesel_configurations = ethanol_biodiesel_configurations.intersection(cellulosic_ethanol_configurations)
conventional_ethanol_biodiesel_configurations = ethanol_biodiesel_configurations.difference(cellulosic_ethanol_configurations)
oil_configurations = biodiesel_configurations.difference(ethanol_configurations)
cellulosic_oil_configurations = oil_configurations.intersection(cellulosic_configurations)

system_factories = {
    -1: create_sugarcane_to_ethanol_system,
    -2: create_sugarcane_to_ethanol_combined_1_and_2g,
    -3: create_sugarcane_to_sugar_and_ethanol_system, # Not yet optimized for sugar production, not recommended to use
    1: create_oilcane_to_biodiesel_and_ethanol_1g,
    2: create_oilcane_to_biodiesel_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation,
    3: create_oilcane_to_crude_oil_and_ethanol_1g,
    4: create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation,
    5: create_oilcane_to_biodiesel_1g,
    6: create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation,
    7: create_oilcane_to_biodiesel_1g,
    8: create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation,
    9: create_oilcane_to_biodiesel_combined_1_and_2g_post_fermentation_oil_separation,
    1j: create_oilcane_to_biodiesel_and_actag_1g,
    2j: create_oilcane_to_biodiesel_and_actag_combined_1_and_2g_post_fermentation_oil_separation,
}
system_factory_options = {
    -1: dict(use_area_convention=True, pellet_bagasse=False, dry_bagasse=True),
    5: dict(fed_batch=True),
    6: dict(fed_batch=True),
    7: dict(fed_batch=False),
    8: dict(fed_batch=False),
    9: dict(fed_batch=False, oil_extraction='screwpress'),
}
area_names = {
    -1: ['Feedstock handling', 'Juicing', 'EtOH prod.', 'CH&P', 'Utilities', 
         'HXN', 'Storage', 'Wastewater treatment'],
    -2: ['Feedstock handling', 'Juicing', 'Pretreatment', 'EtOH prod.',
         'Wastewater treatment', 'CH&P', 'Utilities', 'HXN', 'Storage'],
    -3: ['Feedstock handling', 'Juicing', 'EtOH prod.', 'CH&P', 'Utilities', 
         'HXN', 'Storage', 'Wastewater treatment'],
    1: ['Feedstock handling', 'Juicing', 'EtOH prod.', 'Oil ext.', 'Biod. prod.', 
        'CH&P', 'Utilities', 'HXN', 'Storage', 'Wastewater treatment'],
    2: ['Feedstock handling', 'Juicing', 'Pretreatment', 'EtOH prod.', 'Wastewater treatment',
        'Oil ext.', 'CH&P',  'Biod. prod.', 'Utilities', 'HXN', 'Storage'],
    3: ['Feedstock handling', 'Juicing', 'EtOH prod.', 'Oil ext.',  'CH&P', 
        'Utilities', 'HXN', 'Storage', 'Wastewater treatment'],
    4: ['Feedstock handling', 'Juicing', 'Pretreatment', 'EtOH prod.', 
        'Wastewater treatment', 'Oil ext.', 'CH&P', 'Utilities', 'HXN', 'Storage'],
    5: ['Feedstock handling', 'Juicing', 'Oil prod. & ext.', 'Biod. prod.', 
        'CH&P', 'Utilities', 'HXN', 'Storage', 'Wastewater treatment'],
    6: ['Feedstock handling',  'Juicing', 'Pretreatment', 'Oil prod. & ext.',
        'Wastewater treatment', 'Biod. prod.', 'CH&P', 'Utilities', 'HXN', 'Storage'],
    # 9: ['Feedstock handling', 'Juicing', 'TAG prod.', 'Oil ext.', 'AcTAG sep.', 
    #     'Biod. prod.', 'CH&P', 'Utilities', 'HXN', 'Storage'],
    # 10: ['Feedstock handling', 'Juicing', 'Pretreatment', 'TAG prod.', 'AcTAG sep.', 
    #      'Wastewater treatment', 'Oil ext.', 'CH&P',  'Biod. prod.', 'Utilities', 'HXN', 'Storage'],
}
area_names[7] = area_names[5]
area_names[9] = area_names[8] = area_names[6]

price_distributions_2022 = None
price_distributions_2023 = None
def get_price_distributions_module(year):
    if year == 2022:
        global price_distributions_2022
        if price_distributions_2022 is None: 
            from .data import price_distributions_2022
        return price_distributions_2022
    elif year == 2023:
        global price_distributions_2023
        if price_distributions_2023 is None: 
            from .data import price_distributions_2023
        return price_distributions_2023
    else:
        raise ValueError(f'invalid year {year}; year must be either 2022 or 2023')

def rename_storage_units(units, storage_area):
    bst.rename_units([i for i in units if bst.is_storage_unit(i)], storage_area)

def linear_fit(x0, x1, y0, y1):
    m = (y1 - y0) / (x1 - x0)
    b = y0 - m * x0
    return m, b

def linear_val(x, mb):
    m, b = mb
    return x * m + b

def exponential_fit(x0, x1, y0, y1): # A x ^ n
    x0 = log(x0)
    x1 = log(x1)
    y0 = log(y0)
    y1 = log(y1)
    n, logA = linear_fit(x0, x1, y0, y1)
    return n, exp(logA)

def exponential_val(x, nA): # b e ^ m
    n, A = nA    
    return A * x ** n

def YRCP2023():
    Biorefinery.default_conversion_performance_distribution = 'longterm'
    Biorefinery.default_prices_correleted_to_crude_oil = True
    Biorefinery.default_oil_content_range = [0, 15]
    Biorefinery.default_baseline_oil_content = 1.8
    Biorefinery.default_year = 2023
    Biorefinery.default_WWT = 'high-rate'

class Biorefinery:
    cache = {}
    baseline_dry_biomass_yield = 25.62 # dry MT / ha / yr
    baseline_available_land = 1600000 * 0.3 / baseline_dry_biomass_yield # ha
    set_feedstock_line = set_line_composition_parameters
    default_prices_correleted_to_crude_oil = False
    default_conversion_performance_distribution = 'longterm'
    default_year = 2022
    default_WWT = None
    default_oil_content_range = [5, 15]
    default_baseline_oil_content = 10
    baseline_moisture_content = 0.70
    baseline_feedstock_CF = GWP_characterization_factors['sugarcane'] # [kg CO2-eq / kg sugarcane] with transportation
    baseline_feedstock_price = 0.035 # [USD / kg sugarcane] with transportation
    baseline_transportation_cost = 0.11 * baseline_feedstock_price # Energy cane usage for cellulosic ethanol: estimation of feedstock costs; https://ageconsearch.umn.edu/record/46837/files/Energy%20Cane%20Feedstock%20Estimation_SAEA1.pdf
    baseline_transportation_CF = 0.029 * baseline_feedstock_CF # Estimated from GREET 2020
    
    def update_feedstock(self):
        feedstock = self.feedstock
        feedstock.price = self.feedstock_price
        feedstock.set_CF(GWP, self.feedstock_CF)
        self.sys.rescale(feedstock, self.feedstock_flow / feedstock.F_mass)
    
    @property
    def feedstock_price(self): # [USD / kg] with transportation
        return (
            (self.baseline_feedstock_price 
             - self.baseline_transportation_cost) * self.baseline_dry_biomass_yield / self.dry_biomass_yield
            + self.transportation_cost
        )
    
    @property
    def feedstock_CF(self): # [kg CO2-eq / kg] with transportation
        return (
            (self.baseline_feedstock_CF
             - self.baseline_transportation_CF) * self.baseline_dry_biomass_yield / self.dry_biomass_yield
            + self.transportation_CF
        )
    
    @property
    def transportation_cost(self): # [USD / kg] 
        return (
            self.baseline_transportation_cost
            * sqrt(self.available_land / self.baseline_available_land)
        )
    
    @property
    def transportation_CF(self): # [kg CO2-eq / kg] 
        return (
            self.baseline_transportation_CF
            * sqrt(self.available_land / self.baseline_available_land)
        )
    
    @property
    def chemicals(self):
        try:
            chemicals = self._chemicals
        except AttributeError:
            self._chemicals = chemicals = create_chemicals()
        return chemicals
    
    _derivative_disabled = False
    @classmethod
    def disable_derivative(cls, disable=True):
        cls._derivative_disabled = disable
    @classmethod
    def enable_derivative(cls, enable=True):
        cls._derivative_disabled = not enable
    
    @property
    def feedstock_flow(self): # kg / hr
        annual_crushing_capacity = self.annual_crushing_capacity
        feedstock = self.feedstock
        moisture_content = feedstock.get_mass_fraction('Water')
        dry_content = 1 - moisture_content
        return 1000 * annual_crushing_capacity / self.tea.operating_hours / dry_content
    
    @property
    def annual_crushing_capacity(self): # dry MT / yr
        return self.available_land * self.dry_biomass_yield
    
    @property
    def baseline_annual_crushing_capacity(self): # dry MT / yr
        return self.baseline_available_land * self.baseline_dry_biomass_yield
    
    @property
    def wet_biomass_yield(self):
        dry_biomass_yield = self.dry_biomass_yield
        moisture_content = self.feedstock.get_mass_fraction('Water')
        dry_fraction = 1 - moisture_content
        return dry_biomass_yield / dry_fraction
        
    def use_max_microbial_oil_performance(self):
        max_lipid_yield_glucose = perf.max_lipid_yield_glucose
        max_lipid_yield_xylose = perf.max_lipid_yield_xylose
        self.set_glucose_to_microbial_oil_yield.setter(max_lipid_yield_glucose * 100)
        self.set_xylose_to_microbial_oil_yield.setter(max_lipid_yield_xylose * 100)
        self.set_fermentation_microbial_oil_productivity.setter(1.902) # Similar to sugarcane ethanol
        self.set_fermentation_microbial_oil_titer.setter(137) # Similar to sugarcane ethanol
        
    def __new__(cls, name, chemicals=None, reduce_chemicals=False, 
                 avoid_natural_gas=True, conversion_performance_distribution=None,
                 year=None, cache=cache, feedstock_line=None,
                 prices_correleted_to_crude_oil=None, WWT_kwargs=None,
                 oil_content_range=None):
        if year is None: year = cls.default_year
        if conversion_performance_distribution is None: 
            conversion_performance_distribution = cls.default_conversion_performance_distribution
        else:
            conversion_performance_distribution = conversion_performance_distribution.replace(' ', '').replace('-', '').lower()
        if prices_correleted_to_crude_oil is None:
            prices_correleted_to_crude_oil = cls.default_prices_correleted_to_crude_oil
        number, agile, energycane, = configuration = parse_configuration(name)
        if WWT_kwargs is None:
            if cls.default_WWT is None:
                WWT_key = None
            else:
                WWT_kwargs = dict(kind=cls.default_WWT)
                WWT_key = cls.default_WWT
        else:
            WWT_key = WWT_kwargs.get('kind', None)
        key = (number, agile, energycane, WWT_key, conversion_performance_distribution, year)
        if cache and key in cache: 
            return cache[key]
        else:
            self = super().__new__(cls)
        flowsheet_name = format_configuration(configuration, latex=False)
        flowsheet = bst.Flowsheet(flowsheet_name)
        main_flowsheet.set_flowsheet(flowsheet)
        if chemicals: self._chemicals = chemicals
        else: chemicals = self.chemicals
        bst.settings.set_thermo(chemicals)
        load_process_settings()
        s = flowsheet.stream
        u = flowsheet.unit
        get_stream = lambda ID: s[ID] if ID in flowsheet.stream else MockStream(ID) # Retrieve mock-stream in case stream does not exist in configuration
        
        ## System
        if number in system_factories:
            self.cane_sys = cane_sys = system_factories[number](operating_hours=24 * 200, WWT_kwargs=WWT_kwargs,
                                                                **system_factory_options.get(number, {}), autorename=True)
        else:
            raise NotImplementedError(number)
        cane_sys.set_tolerance(rmol=1e-5, mol=1e-2, subsystems=True, subfactor=1.5)
        
        if number == 1:
            cane_sys.prioritize_unit(u.T508)
        elif number == 2:
            cane_sys.prioritize_unit(u.T808)
        
        if number < 0:
            s.sugarcane.register_alias('feedstock')
        else:
            s.oilcane.register_alias('feedstock')
        feedstock = s.feedstock
        
        chemicals.define_group(
            name='Fiber',
            IDs=['Cellulose', 'Hemicellulose', 'Lignin'],
            composition=[0.4704 , 0.2775, 0.2520],
            wt=True, # Composition is given as weight
        )
        chemicals.define_group(
            name='Sugar',
            IDs=['Sucrose', 'Glucose'],
            # Default composition as equimolar
        )
        
        if energycane:
            feedstock.reset_flow(**energycane_dct, total_flow=feedstock.F_mass, units='kg/hr')
            feedstock.ID = 'energycane'
        
        BT = flowsheet(bst.BoilerTurbogenerator)
        if number not in cellulosic_configurations: BT.boiler_efficiency = 0.89
        
        ## Unit groups
        areas = area_names[number]
        rename_storage_units(cane_sys.units, (areas.index('Storage') + 1) * 100)
        unit_groups = UnitGroup.group_by_area(cane_sys.units)
        for i, j in zip(unit_groups, areas): i.name = j
        for i in unit_groups: i.autofill_metrics(shorthand=True)
        
        for HXN_group in unit_groups:
            if HXN_group.name == 'HXN':
                HXN_group.filter_savings = False # Allow negative values in heat utilities
                HXN = HXN_group.units[0]
                try:
                    assert isinstance(HXN, bst.HeatExchangerNetwork)
                except: breakpoint()
        HXN.raise_energy_balance_error = False
        HXN.vle_quenched_streams = False
        
        ## Split bagasse to boiler meet energy demand
        if avoid_natural_gas:
            for splitter in flowsheet.unit:
                if getattr(splitter, 'isbagasse_splitter', False): break
            else:
                avoid_natural_gas = False
        if avoid_natural_gas:
            cane_sys.add_specification(self.update_feedstock)
            self.bagasse_splitter = splitter
            minimum_fraction_burned = 0
            maximum_fraction_burned = 0.7
            self.recycle_data = recycle_data = {}
            @cane_sys.add_bounded_numerical_specification(
                x0=minimum_fraction_burned, x1=maximum_fraction_burned, 
                xtol=1e-4, ytol=100, args=(splitter,)
            )
            def adjust_bagasse_to_boiler(fraction_burned, splitter):
                # Returns energy consumption at given fraction processed (not sent to boiler).
                splitter.split[:] = 1 - fraction_burned
                operation_mode = getattr(sys, 'active_operation_mode', None)
                if fraction_burned in (minimum_fraction_burned, maximum_fraction_burned):
                    key = (operation_mode, fraction_burned)
                else:
                    key = (operation_mode, 'last')
                if key in recycle_data: 
                    material_data = recycle_data[key]
                else:
                    recycle_data[key] = material_data = cane_sys.get_material_data()
                try:
                    cane_sys.simulate(material_data=material_data, update_material_data=True)
                except:
                    cane_sys.simulate()
                excess = BT._excess_electricity_without_natural_gas
                if fraction_burned == minimum_fraction_burned and excess > 0:
                    splitter.neglect_natural_gas_streams = False # No need to neglect
                    return 0 # No need to burn bagasse
                elif fraction_burned == maximum_fraction_burned and excess < 0: 
                    splitter.neglect_natural_gas_streams = False # Cannot be neglected
                    return 0 # Cannot satisfy energy demand even at 30% sent to boiler (or minimum fraction processed)
                else:
                    splitter.neglect_natural_gas_streams = True
                    return excess
            @cane_sys.add_specification(args=(splitter,))
            def assume_negligible_natural_gas_streams(splitter):
                if splitter.neglect_natural_gas_streams:
                    for i in natural_gas_streams: i.empty()
        else:
            cane_sys.add_specification(self.update_feedstock, simulate=True)
        
        if abs(number) in cellulosic_configurations:
            prs = flowsheet(bst.PretreatmentReactorSystem)
            saccharification = flowsheet(bst.Saccharification)
            seed_train = flowsheet(bst.SeedTrain)
            try:
                fermentor = flowsheet(bst.CoFermentation)
            except:
                fermentor = flowsheet(bst.AeratedBioreactor)
            self.pretreatment_rxnsys = tmo.ReactionSystem(
                prs.reactions, saccharification.saccharification
            )
            self.fermentation_rxnsys = tmo.ReactionSystem(
                seed_train.reactions, fermentor.cofermentation
            )
            self.cellulosic_rxnsys = tmo.ReactionSystem(
                prs.reactions, saccharification.saccharification,
                seed_train.reactions, fermentor.cofermentation
            )
            saccharification.saccharification.X[0] = 0.0 # Baseline
            prs.reactions.X[10] = 0.0 # baseline
        else:
            try:
                fermentor = flowsheet(bst.Fermentation)
            except:
                fermentor = flowsheet(bst.AeratedBioreactor)
        self.fermentor = fermentor
        
        def set_glucose_yield(glucose_yield):
            if number in cellulosic_configurations:
                glucose_yield *= 0.01
                X1 = prs.reactions.X[0]
                X1_side = prs.reactions.X[1:3].sum()
                X2_side = saccharification.saccharification.X[:2].sum()
                saccharification.saccharification.X[2] = X2 = (glucose_yield - X1) / (1 - X1_side)
                X_excess = (X2_side + X2) - 1
                if X_excess > 0: breakpoint()
                
        def set_xylose_yield(xylose_yield):
            if number in cellulosic_configurations:
                xylose_yield *= 0.01
                X1_side = prs.reactions.X[9:11].sum()
                prs.reactions.X[8] = X1 = xylose_yield
                X_excess = (X1_side + X1) - 1
                if X_excess > 0.: breakpoint()
        
        if agile: 
            self.oilsorghum = oilsorghum = bst.Stream(
                ID='oilsorghum', phase='l', T=298.15, P=101325, 
                Water=2.333e+05, Glucose=3703, Sucrose=4.196e+04, Ash=2000, 
                Cellulose=2.227e+04, Hemicellulose=1.314e+04, Lignin=1.193e+04, 
                Solids=5000, units='kg/hr'
            )
            
            sys = bst.AgileSystem()
            @sys.operation_parameter(mode_dependent=True)
            def set_oil_content(oil_content, mode):
                F_mass = feedstock.F_mass
                feedstock.copy_flow(mode.feedstock)
                feedstock.F_mass = F_mass
                if number > 0: 
                    set_oil_fraction(oil_content, feedstock,
                                     FFA_fraction=mode.FFA_content,
                                     z_mass_carbs_baseline=mode.z_mass_carbs_baseline,
                                     PL_fraction=mode.PL_content)
            sys.operation_parameter(set_glucose_yield)
            sys.operation_parameter(set_xylose_yield)
            
            if energycane:
                self.cane_mode = cane_mode = sys.operation_mode(cane_sys,
                    operating_hours=180*24, oil_content=0.02, feedstock=feedstock.copy(),
                    z_mass_carbs_baseline=0.091,
                    z_mass_solids_baseline=0., 
                    z_mass_ash_baseline=0.028,
                    z_mass_water_baseline=0.60,
                    glucose_yield=85, xylose_yield=65, 
                    FFA_content=0.10, PL_content=0.10
                )
            else:
                self.cane_mode = cane_mode = sys.operation_mode(cane_sys,
                    operating_hours=180*24, oil_content=0.10, feedstock=feedstock.copy(),
                    z_mass_carbs_baseline=0.1491, glucose_yield=85, xylose_yield=65, 
                    FFA_content=0.10, PL_content=0.10
                )
            self.sorghum_mode = sorghum_mode = sys.operation_mode(cane_sys, 
                operating_hours=60*24, oil_content=0.07, glucose_yield=79, xylose_yield=86,
                feedstock=oilsorghum,
                z_mass_carbs_baseline=0.1371, FFA_content=0.10, PL_content=0.10,
            )
            tea = create_tea(sys)
            tea.operating_days = 260 
            tea.IRR = 0.10
        else:
            sys = cane_sys
            tea = create_tea(sys)
            tea.operating_days = 200
            tea.IRR = 0.10
        
        tea.income_tax = 0.21 # Davis et al. 2018; https://www.nrel.gov/docs/fy19osti/71949.pdf
        
        ## Specifications for analysis
        self.composition_specification = composition_specification = CaneCompositionSpecification(feedstock)
        if number < 0:
            oil_extraction_specification = MockExtractionSpecification()
        else:
            crushing_mill = flowsheet(bst.CrushingMill) # Separates bagasse
            _, screen = flowsheet(bst.VibratingScreen)
            juice = screen.outs[0]
            if number in cellulosic_configurations:
                pressure_filter = flowsheet(bst.PressureFilter) # Separates lignin
            else:
                pressure_filter  = None
            if number in screwpress_microbial_oil_recovery:
                screw_press = flowsheet(bst.ScrewPress) # Separates oil from cell mass
            else:
                screw_press = None
            if number in screwpress_microbial_oil_recovery:
                cellmass_centrifuge = flowsheet(bst.SolidLiquidsSplitCentrifuge) # Separates cell mass, oil, and water
            else:
                cellmass_centrifuge = flowsheet(bst.SolidsCentrifuge) # Separates cell mass
            if isinstance(cellmass_centrifuge, list):
                cellmass_centrifuge = sorted([i for i in cellmass_centrifuge if type(i) is bst.SolidsCentrifuge], key=lambda x: x.ID)[0]
            cellmass_centrifuge.strict_moisture_content = False
            oil_extraction_specification = OilExtractionSpecification(
                sys, crushing_mill, pressure_filter, cellmass_centrifuge, juice, screw_press
            )
        
        ## Account for cellulosic vs advanced RINs
        if number in cellulosic_ethanol_configurations:
            # Note that GREET cellulosic ethanol from corn stover results in a 
            # GWP of 0.41 kg CO2-eq / L-ethanol. So the cellulosic ethanol from
            # bagasse (~0.34, 0.32 for configurations S2 and O2) can certainly
            # apply as an EPA cellulosic ethanol pathway.
            RIN_splitter = bst.Splitter('RIN_splitter',
                ins=s.ethanol,
                outs=['advanced_ethanol', 'cellulosic_ethanol'],
                split=0.5
            )
            @RIN_splitter.add_specification(run=True)
            def adjust_ethanol_split():
                # outs: stream sequence
                # [0] Advanced biofuel ethanol
                # [1] Cellulosic biofuel ethanol
                glucose_yield = fermentor.cofermentation[0].product_yield('Ethanol', 'wt')
                xylose_yield = fermentor.cofermentation[1].product_yield('Ethanol', 'wt')
                juice_sugar = s.juice.imass['Glucose', 'Sucrose'].sum() * glucose_yield
                hydrolysate_sugar = (
                    s.slurry.imass['Glucose'] * glucose_yield
                    + s.slurry.imass['Xylose', 'Arabinose'].sum() * xylose_yield
                )
                RIN_splitter.split[:] = juice_sugar / (juice_sugar + hydrolysate_sugar)
            
            cane_sys.update_configuration([*cane_sys.units, RIN_splitter])
            assert RIN_splitter in cane_sys.units
        elif number in conventional_ethanol_configurations:
            s.ethanol.ID = 'advanced_ethanol'
            s.advanced_ethanol.register_alias('ethanol')
        elif number in conventional_ethanol_configurations:
            s.ethanol.ID = 'advanced_ethanol'
            s.advanced_ethanol.register_alias('ethanol')
        if number in biodiesel_configurations:
            if (number not in cellulosic_ethanol_configurations
                and number in cellulosic_configurations):
                # Note that GREET cellulosic ethanol from corn stover results in a 
                # GWP of 0.41 kg CO2-eq / L-ethanol. So the cellulosic ethanol from
                # bagasse (~0.34, 0.32 for configurations S2 and O2) can certainly
                # apply as an EPA cellulosic ethanol pathway.
                RIN_splitter = bst.Splitter('RIN_splitter',
                    ins=s.biodiesel,
                    outs=['biomass_based_diesel', 'cellulosic_based_diesel'],
                    split=0.5
                )
                @RIN_splitter.add_specification(run=True)
                def adjust_biodiesel_split():
                    # outs: stream sequence
                    # [0] Biomass based diesel
                    # [1] Cellulosic based diesel
                    glucose_yield = fermentor.cofermentation[0].product_yield('TriOlein', 'wt')
                    xylose_yield = fermentor.cofermentation[1].product_yield('TriOlein', 'wt')
                    juice_sugar = s.juice.imass['Glucose', 'Sucrose'].sum() * glucose_yield
                    hydrolysate_sugar = (
                        s.slurry.imass['Glucose'] * glucose_yield
                        + s.slurry.imass['Xylose', 'Arabinose'].sum() * xylose_yield
                    )
                    RIN_splitter.split[:] = juice_sugar / (juice_sugar + hydrolysate_sugar)
                
                cane_sys.update_configuration([*cane_sys.units, RIN_splitter])
                assert RIN_splitter in cane_sys.units
            else:
                # A biodiesel stream should already exist
                s.biodiesel.register_alias('biomass_based_diesel')
        
        ## Additional modifications for speed
        for i in sys.units:
            if isinstance(i, bst.MultiEffectEvaporator): i.flash = False
        HXN.acceptable_energy_balance_error = 0.02
        HXN.replace_unit_heat_utilities = False
        HXN.force_ideal_thermo = True
        HXN.cache_network = True
        HXN.avoid_recycle = True
        try: AD = flowsheet(bst.AerobicDigestion)
        except: pass
        else:
            WWTsys = cane_sys.find_system(AD)
            WWTsys.set_tolerance(mol=10, method='fixed-point')
        
        ## LCA
        # Set non-negligible characterization factors
        stream_names = ('FGD_lime', 'cellulase', 'urea', 'caustic', 'cellulosic_ethanol',
                        'catalyst', 'methanol', 'HCl', 'NaOH', 'crude_glycerol', 'pure_glycerine',
                        'denaturant', 'ethanol', 'advanced_ethanol', 'cellulosic_ethanol',
                        'biomass_based_diesel', 'cellulosic_based_diesel', 'natural_gas',
                        'biodiesel', 'lime', 'H3PO4')
        (FGD_lime, cellulase, urea, caustic, cellulosic_ethanol,
         catalyst, methanol, HCl, NaOH, crude_glycerol, pure_glycerine,
         denaturant, ethanol, advanced_ethanol, cellulosic_ethanol,
         biomass_based_diesel, cellulosic_based_diesel, natural_gas, 
         biodiesel, lime, H3PO4) = [get_stream(i) for i in stream_names]
        
        def get_renamed_stream(search, ID):
            if search in flowsheet.stream:
                stream = s[search]  
                stream.ID = ID
            else:
                stream = MockStream(ID)
            return stream
        
        NaOCl = get_renamed_stream('naocl_R602', 'NaOCl')
        citric_acid = get_renamed_stream('citric_R602', 'citric_acid')
        bisulfite = get_renamed_stream('bisulfite_R602', 'bisulfite')
        polishing_filter_air = get_renamed_stream('air_R603', 'polishing_filter_air')
        polishing_filter_vent = get_renamed_stream('vent_R603', 'polishing_filter_vent')
        set_GWPCF(NaOCl, 'NaOCl')
        set_GWPCF(citric_acid, 'citric acid')
        set_GWPCF(bisulfite, 'bisulfite')
        set_GWPCF(feedstock, 'sugarcane')
        set_GWPCF(H3PO4, 'H3PO4')
        set_GWPCF(lime, 'lime', dilution=0.046) # Diluted with water
        set_GWPCF(denaturant, 'gasoline')
        set_GWPCF(FGD_lime, 'lime', dilution=0.451)
        set_GWPCF(cellulase, 'cellulase', dilution=0.05) 
        set_GWPCF(urea, 'urea')
        set_GWPCF(caustic, 'NaOH', 0.5)
        set_GWPCF(catalyst, 'methanol catalyst mixture')
        set_GWPCF(methanol, 'methanol')
        set_GWPCF(HCl, 'HCl')
        set_GWPCF(NaOH, 'NaOH')
        set_GWPCF(pure_glycerine, 'pure-glycerol')
        set_GWPCF(crude_glycerol, 'crude-glycerol', dilution=0.80)
        set_GWPCF(biodiesel, 'biodiesel displacement')
        bst.PowerUtility.set_CF(GWP, GWP_characterization_factors['Electricity'])
        self.natural_gas_streams = natural_gas_streams = [natural_gas]
        for stream in natural_gas_streams: set_GWPCF(stream, 'CH4')
        
        ## Add BioSTEAM objects to module for easy access
        dist = get_price_distributions_module(year)
        
        ## Model
        model = bst.Model(sys, exception_hook='raise', retry_evaluation=False)
        parameter = model.parameter
        metric = model.metric
        
        # DO NOT DELETE:
        # natural_gas.phase = 'g'
        # natural_gas.set_property('T', 60, 'degF')
        # natural_gas.set_property('P', 14.73, 'psi')
        # original_value = natural_gas.imol['CH4']
        # natural_gas.imass['CH4'] = 1 
        # V_ng = natural_gas.get_total_flow('m3/hr')
        # natural_gas.imol['CH4'] = original_value
        V_ng = 1.473318463076884 # Natural gas volume at 60 F and 14.73 psi [m3 / kg]
        
        def potential_shortterm_gain_in_performance(lb, ub, *args, **kwargs):
            distribution = shape.Uniform(lb, lb + 0.15 * (ub - lb))
            return parameter(*args, distribution=distribution, bounds=(lb, ub), baseline=lb, **kwargs)
        
        def potential_longterm_gain_in_performance(lb, ub, *args, **kwargs):
            distribution = shape.Uniform(lb, ub)
            return parameter(*args, distribution=distribution, bounds=(lb, ub), baseline=lb, **kwargs)
        
        def uniform(lb, ub, *args, **kwargs):
            return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub), **kwargs)
        
        def default(baseline, *args, **kwargs):
            lb = 0.75*baseline
            ub = 1.25*baseline
            return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub),
                             baseline=baseline, **kwargs)
        
        def default_gwp(baseline, *args, **kwargs):
            lb = 0.90*baseline
            ub = 1.10*baseline
            return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub), 
                             baseline=baseline, **kwargs)
        
        def triangular(lb, mid, ub, *args, **kwargs):
            return parameter(*args, distribution=shape.Triangle(lb, mid, ub), bounds=(lb, ub), **kwargs)
        
        if oil_content_range is None:
            oil_content_range = self.default_oil_content_range
        
        if conversion_performance_distribution == 'shortterm':
            performance = potential_shortterm_gain_in_performance
        elif conversion_performance_distribution == 'longterm':
            performance = potential_longterm_gain_in_performance
        else:
            raise ValueError("`conversion_performance_distribution` must be either 'longterm' or 'shortterm")
            
        @performance(60, 90, units='%', kind='coupled')
        def set_juicing_oil_recovery(juicing_oil_recovery):
            oil_extraction_specification.load_juicing_oil_recovery(juicing_oil_recovery / 100.)
        
        @performance(70, 90, units='%', kind='coupled')
        def set_microbial_oil_recovery(microbial_oil_recovery):
            oil_extraction_specification.load_microbial_oil_recovery(microbial_oil_recovery / 100.)
        
        @performance(70.0, 90, units='%', kind='coupled')
        def set_bagasse_oil_recovery(bagasse_oil_recovery):
            oil_extraction_specification.load_bagasse_oil_recovery(bagasse_oil_recovery / 100.)
    
        # Baseline from Huang's 2016 paper, but distribution more in line with Florida sugarcane harvesting (3-5 months)
        @uniform(4 * 30, 6 * 30, units='day/yr', baseline=180)
        def set_cane_operating_days(cane_operating_days):
            if agile:
                cane_mode.operating_hours = cane_operating_days * 24
            else:
                tea.operating_days = cane_operating_days
    
        # From Ed Cahoon and Huang 2017
        @uniform(30, 60, units='day/yr', baseline=45)
        def set_sorghum_operating_days(sorghum_operating_days):
            if agile: sorghum_mode.operating_hours = sorghum_operating_days * 24
        
        @default(self.baseline_available_land, units='ha / yr', kind='isolated')
        def set_available_land(available_land):
            self.available_land = available_land
    
        @parameter(distribution=dist.copd, element='crude oil', baseline=dist.mcop, units='USD/barrel')
        def set_crude_oil_price(price):
            self.crude_oil_price = price
        
        @default(self.baseline_feedstock_price, element='feedstock', units='USD/kg')
        def set_baseline_feedstock_price(price):
            self.baseline_feedstock_price = price
        
        if prices_correleted_to_crude_oil:
            predict = lambda name, scalar: dist.models[name].predict(np.array([[scalar]]))[0]
                                                                     
            @parameter(distribution=dist.offsets['Cellulosic ethanol'],
                       element='Cellulosic ethanol', baseline=0., units='USD/L')
            def set_cellulosic_ethanol_price(price): 
                cellulosic_ethanol.price = predict('Cellulosic ethanol', self.crude_oil_price + price) * ethanol_L_per_kg
                
            @parameter(distribution=dist.offsets['Advanced ethanol'],
                       element='Advanced ethanol', baseline=0., units='USD/L')
            def set_advanced_ethanol_price(price): 
                advanced_ethanol.price =  predict('Advanced ethanol', self.crude_oil_price + price) * ethanol_L_per_kg
                
            # USDA ERS historical price data
            @parameter(distribution=dist.offsets['Biomass based diesel'], 
                       element='Biomass based diesel', units='USD/L', baseline=0.)
            def set_biomass_based_diesel_price(price):
                biomass_based_diesel.price =  predict('Biomass based diesel', self.crude_oil_price + price) * biodiesel_L_per_kg
        
            @parameter(distribution=dist.offsets['Cellulosic based diesel'],
                       element='Cellulosic based diesel', units='USD/L', baseline=0.)
            def set_cellulosic_based_diesel_price(price):
                cellulosic_based_diesel.price =  predict('Cellulosic based diesel', self.crude_oil_price + price) * biodiesel_L_per_kg
        
            # https://www.eia.gov/energyexplained/natural-gas/prices.php
            @parameter(distribution=dist.offsets['Natural gas'],
                       element='Natural gas', units='USD/m3', baseline=0.)
            def set_natural_gas_price(price): 
                BT.natural_gas_price =  predict('Natural gas', self.crude_oil_price + price) * V_ng
        
            @parameter(distribution=dist.offsets['Electricity'], element='electricity',
                       units='USD/kWhr', baseline=0.)
            def set_electricity_price(price): 
                bst.PowerUtility.price = predict('Electricity', self.crude_oil_price + price)
                
        else:
            # USDA ERS historical price data with EPA RIN prices
            @parameter(distribution=dist.cepd, element='Cellulosic ethanol', 
                       baseline=dist.mcep, units='USD/L')
            def set_cellulosic_ethanol_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to Nov 2020
                cellulosic_ethanol.price = price * ethanol_L_per_kg
                
            @parameter(distribution=dist.aepd, element='Advanced ethanol', 
                       baseline=dist.maep, units='USD/L')
            def set_advanced_ethanol_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to Nov 2020
                advanced_ethanol.price = price * ethanol_L_per_kg
                
            # USDA ERS historical price data
            @parameter(distribution=dist.bpd, element='Biomass based diesel', units='USD/L', baseline=dist.mbp)
            def set_biomass_based_diesel_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to March 2021
                biomass_based_diesel.price = price * biodiesel_L_per_kg
        
            @parameter(distribution=dist.cbpd, element='Cellulosic based diesel', units='USD/L', baseline=dist.mcbp)
            def set_cellulosic_based_diesel_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to March 2021
                cellulosic_based_diesel.price = price * biodiesel_L_per_kg
        
            # https://www.eia.gov/energyexplained/natural-gas/prices.php
            @parameter(distribution=dist.natural_gas_price_distribution, element='Natural gas', units='USD/m3',
                       baseline=4.73 * 35.3146667/1e3)
            def set_natural_gas_price(price): 
                BT.natural_gas_price = price * V_ng
        
            @parameter(distribution=dist.electricity_price_distribution, units='USD/kWhr',
                       element='electricity', baseline=dist.mean_electricity_price)
            def set_electricity_price(price): 
                bst.PowerUtility.price = price
        
        # 10% is suggested for waste reducing, but 15% is suggested for investment
        @uniform(10., 15., units='%', baseline=10)
        def set_IRR(IRR):
            tea.IRR = IRR / 100.
        
        @uniform(0.10, 0.22, units='USD/kg', element=crude_glycerol.ID)
        def set_crude_glycerol_price(price):
            crude_glycerol.price = price
    
        @default(0.65, units='USD/kg', element=pure_glycerine.ID)
        def set_pure_glycerol_price(price):
            pure_glycerine.price = price
        
        @default(72, units='hr', element='Saccharification')
        def set_saccharification_reaction_time(reaction_time):
            if abs(number) in cellulosic_configurations: saccharification.tau = reaction_time
        
        @default(0.212, units='USD/kg', element=cellulase.ID)
        def set_cellulase_price(price):
            cellulase.price = price
    
        if abs(number) in cellulosic_configurations:
            cellulase_mixer, = [i for i in flowsheet.unit if hasattr(i, 'enzyme_loading')]
       
        @default(0.02, units='wt. % cellulose', element='cellulase', kind='coupled')
        def set_cellulase_loading(cellulase_loading):
            if abs(number) in cellulosic_configurations: cellulase_mixer.enzyme_loading = cellulase_loading
        
        @default(PRS_cost_item.cost, units='million USD', element='Pretreatment reactor system')
        def set_reactor_base_cost(base_cost):
            PRS_cost_item.cost = base_cost
        
        @performance(85, 97.5, units='%', element='Pretreatment and saccharification', kind='coupled')
        def set_cane_glucose_yield(cane_glucose_yield):
            if agile:
                cane_mode.glucose_yield = cane_glucose_yield
            elif abs(number) in cellulosic_configurations:
                set_glucose_yield(cane_glucose_yield)
        
        @uniform(79, 97.5, units='%', element='Pretreatment and saccharification',
                 baseline=79, kind='coupled')
        def set_sorghum_glucose_yield(sorghum_glucose_yield):
            if not agile: return
            sorghum_mode.glucose_yield = sorghum_glucose_yield
            
        @performance(65, 97.5, units='%', element='Pretreatment and saccharification', kind='coupled')
        def set_cane_xylose_yield(cane_xylose_yield):
            if agile:
                cane_mode.xylose_yield = cane_xylose_yield
            elif abs(number) in cellulosic_configurations:
                set_xylose_yield(cane_xylose_yield)
        
        @performance(86, 97.5, units='%', element='Pretreatment and saccharification', kind='coupled')
        def set_sorghum_xylose_yield(sorghum_xylose_yield):
            if not agile: return
            sorghum_mode.xylose_yield = sorghum_xylose_yield
        
        @performance(90, 95, units='%', element='Cofermenation', kind='coupled')
        def set_glucose_to_ethanol_yield(glucose_to_ethanol_yield):
            if number in cellulosic_ethanol_configurations:
                glucose_to_ethanol_yield *= 0.01
                # fermentor.cofermentation[2].X = 0.004 # Baseline
                # fermentor.cofermentation[3].X = 0.006 # Baseline
                # fermentor.loss[0].X = 0.03 # Baseline
                split = u.S401.split.mean()
                X1 = split * seed_train.reactions.X[0]
                X2 = split * seed_train.reactions.X[2]
                X3 = (glucose_to_ethanol_yield - X1) / (1 - X1 - X2)
                X_excess = X3 * 1.0526 - 1
                if X_excess > 0.: breakpoint()
                fermentor.cofermentation.X[0] = X3
                fermentor.cofermentation.X[2] = X3 * 0.0526 # 95% towards ethanol, the other 5% goes towards cell mass
        
        @performance(50, 95, units='%', element='Cofermenation', kind='coupled')
        def set_xylose_to_ethanol_yield(xylose_to_ethanol_yield):
            if number in cellulosic_ethanol_configurations:
                xylose_to_ethanol_yield *= 0.01
                split = u.S401.split.mean()
                X1 = split * seed_train.reactions.X[1]
                X2 = split * seed_train.reactions.X[3]
                X3 = (xylose_to_ethanol_yield - X1) / (1 - X1 - X2)
                X_excess = X3 * 1.0526 - 1
                if X_excess > 0.: breakpoint()
                fermentor.cofermentation.X[1] = X3
                fermentor.cofermentation.X[3] = X3 * 0.0526 # 95% towards ethanol, the other 5% goes towards cell mass
    
        @performance(68.5, 137, units='g/L', element='Cofermentation', kind='coupled')
        def set_cofermentation_ethanol_titer(ethanol_titer):
            if number in cellulosic_ethanol_configurations: fermentor.titer = ethanol_titer
    
        @performance(0.951, 1.902, units='g/L', element='Cofermentation')
        def set_cofermentation_ethanol_productivity(ethanol_productivity):
            if number in cellulosic_ethanol_configurations: fermentor.productivity = ethanol_productivity
    
        fed_batch = system_factory_options.get(number, {}).get('fed_batch')
        if fed_batch:
            lipid_yield = perf.fed_batch_lipid_yield_mean
            titer = perf.fed_batch_titer_mean
            productivity = perf.fed_batch_productivity_mean
        else:
            lipid_yield = perf.batch_lipid_yield_mean
            titer = perf.batch_titer_mean
            productivity = perf.batch_productivity_mean
        
        hydrolysate_lipid_yield = perf.hydrolysate_lipid_yield
        hydrolysate_titer = perf.hydrolysate_titer
        hydrolysate_productivity = perf.hydrolysate_productivity
            
        @performance(hydrolysate_lipid_yield * 100, lipid_yield * 100, units='%', element='Cofermenation', kind='coupled')
        def set_glucose_to_microbial_oil_yield(glucose_to_microbial_oil_yield):
            if number in cellulosic_oil_configurations:
                fermentation_reaction = fermentor.cofermentation[0]
                cell_growth_reaction = fermentor.cofermentation[2]
            elif number in oil_configurations:
                fermentation_reaction = fermentor.fermentation_reaction
                cell_growth_reaction = fermentor.cell_growth_reaction
            else:
                return
            glucose_to_microbial_oil_yield *= 0.01
            fermentation_reaction.product_yield(
                product='TAG', product_yield=glucose_to_microbial_oil_yield, basis='wt'
            )
            # Almost all the rest goes towards cell mass and CO2
            if number in cellulosic_oil_configurations:
                # Reaction in parallel
                cell_growth_reaction.X = 0.999 - fermentation_reaction.X
            else:
                # Reaction in series
                cell_growth_reaction.X = 0.999 
            
        @performance(hydrolysate_lipid_yield * 100, lipid_yield * 100, units='%', element='Cofermenation', kind='coupled')
        def set_xylose_to_microbial_oil_yield(xylose_to_microbial_oil_yield):
            if number in cellulosic_oil_configurations:
                xylose_to_microbial_oil_yield *= 0.01
                fermentation_reaction = fermentor.cofermentation[1]
                cell_growth_reaction = fermentor.cofermentation[3]
                fermentation_reaction.product_yield(
                    product='TAG', product_yield=xylose_to_microbial_oil_yield, basis='wt'
                )
                cell_growth_reaction.X = 0.999 - fermentation_reaction.X # Almost all the rest goes towards cell mass and CO2
    
        @performance(hydrolysate_titer, titer, units='g/L', element='Fermentation', kind='coupled')
        def set_fermentation_microbial_oil_titer(microbial_oil_titer):
            if number in oil_configurations: fermentor.titer = microbial_oil_titer
    
        @performance(hydrolysate_productivity, productivity, units='g/L', element='Fermentation')
        def set_fermentation_microbial_oil_productivity(microbial_oil_productivity):
            if number in oil_configurations: fermentor.productivity = microbial_oil_productivity
    
        @default(10, element='oilcane', units='% oil', kind='coupled')
        def set_cane_PL_content(cane_PL_content):
            if agile: cane_mode.PL_content = cane_PL_content / 100.
            else: composition_specification.PL = cane_PL_content / 100.
        
        @default(10, element='oilsorghum', units='% oil', kind='coupled')
        def set_sorghum_PL_content(sorghum_PL_content):
            if agile: sorghum_mode.PL_content = sorghum_PL_content / 100.
        
        @default(10, element='oilcane', units='% oil', kind='coupled')
        def set_cane_FFA_content(cane_FFA_content):
            if agile: cane_mode.FFA_content = cane_FFA_content / 100.
            else: composition_specification.FFA = cane_FFA_content / 100.
        
        @default(10, element='oilsorghum', units='% oil', kind='coupled')
        def set_sorghum_FFA_content(sorghum_FFA_content):
            if agile: sorghum_mode.FFA_content = sorghum_FFA_content / 100. 
    
        @uniform(*oil_content_range, element='oilcane', units='dry wt. %', kind='coupled')
        def set_cane_oil_content(cane_oil_content):
            if number < 0: return
            if agile:
                cane_mode.oil_content = cane_oil_content / 100.
            elif energycane:
                composition_specification.load_oil_content(
                    cane_oil_content / 100.,
                    z_mass_carbs_baseline=0.091,
                    z_mass_solids_baseline=0., 
                    z_mass_ash_baseline=0.028,
                    z_mass_water_baseline=0.60,
                )
            else:
                composition_specification.load_oil_content(
                    cane_oil_content / 100.
                )
    
        @uniform(-3., 0., element='oilsorghum', units='dry wt. %', kind='coupled',
                 baseline=0.)
        def set_relative_sorghum_oil_content(relative_sorghum_oil_content):
            if number < 0: return
            if agile:
                sorghum_mode.oil_content = cane_mode.oil_content + relative_sorghum_oil_content / 100.
    
        @default(23, units='% oil', kind='coupled', name='TAG to FFA conversion')
        def set_TAG_to_FFA_conversion(TAG_to_FFA_conversion):
            if number in conventional_ethanol_biodiesel_configurations:
                u.R301.oil_reaction.X[0] = TAG_to_FFA_conversion / 100.
            elif number in cellulosic_ethanol_biodiesel_configurations:
                u.R401.oil_reaction.X[0] = TAG_to_FFA_conversion / 100.
        
        @default_gwp(feedstock.characterization_factors[GWP], name='GWP', 
                 element='sugarcane', units='kg*CO2-eq/kg')
        def set_feedstock_GWP(value):
            self.baseline_feedstock_CF = value
        
        @default_gwp(methanol.characterization_factors[GWP], name='GWP', 
                     element=methanol.ID, units='kg*CO2-eq/kg')
        def set_methanol_GWP(value):
            methanol.characterization_factors[GWP] = value
        
        # @default(crude_glycerol.characterization_factors[GWP], name='GWP', 
        #          element=crude_glycerol, units='kg*CO2-eq/kg')
        # def set_crude_glycerol_GWP(value):
        #     crude_glycerol.characterization_factors[GWP] = value
        
        @default_gwp(pure_glycerine.characterization_factors[GWP], name='GWP', 
                     element=pure_glycerine.ID, units='kg*CO2-eq/kg')
        def set_pure_glycerine_GWP(value):
            pure_glycerine.characterization_factors[GWP] = value
        
        @default_gwp(cellulase.characterization_factors[GWP], name='GWP', 
                     element=cellulase.ID, units='kg*CO2-eq/kg')
        def set_cellulase_GWP(value):
            cellulase.characterization_factors[GWP] = value * 0.02
        
        @default_gwp(natural_gas.characterization_factors[GWP], name='GWP', 
                     element=natural_gas.ID, units='kg*CO2-eq/kg')
        def set_natural_gas_GWP(value):
            for ng in natural_gas_streams:
                ng.characterization_factors[GWP] = value
        
        @uniform(21., 28., baseline=21., name='Income tax', units='%')
        def set_income_tax(income_tax):
            tea.income_tax = income_tax * 0.01
        
        if agile:
            feedstock_flow = lambda: sys.flow_rates[feedstock] / kg_per_MT # MT / yr
            biodiesel_flow = lambda: (sys.flow_rates.get(cellulosic_based_diesel, 0.) + sys.flow_rates.get(biomass_based_diesel, 0.)) * biodiesel_L_per_kg # L / yr
            ethanol_flow = lambda: (sys.flow_rates.get(cellulosic_ethanol, 0.) + sys.flow_rates.get(advanced_ethanol, 0.)) * ethanol_L_per_kg # L / yr
            natural_gas_flow = lambda: sum([sys.flow_rates[i] for i in natural_gas_streams]) * V_ng # m3 / yr
            crude_glycerol_flow = lambda: sys.flow_rates.get(crude_glycerol, 0.) # kg / yr
            
            @sys.operation_metric(annualize=True)
            def direct_nonbiogenic_emissions(mode):
                return sum([i.F_mol for i in natural_gas_streams]) * chemicals.CO2.MW
            
        else:
            feedstock_flow = lambda: sys.operating_hours * feedstock.F_mass / kg_per_MT # MT / yr
            biodiesel_flow = lambda: sys.operating_hours * biodiesel.F_mass * biodiesel_L_per_kg # L / yr
            ethanol_flow = lambda: sys.operating_hours * ethanol.F_mass * ethanol_L_per_kg # L / yr
            crude_glycerol_flow = lambda: sys.operating_hours * crude_glycerol.F_mass # kg / yr
            natural_gas_flow = lambda: sum([i.F_mass for i in natural_gas_streams]) * sys.operating_hours * V_ng # m3 / yr
            direct_nonbiogenic_emissions = lambda: sum([i.F_mol for i in natural_gas_streams]) * chemicals.CO2.MW * sys.operating_hours
        electricity = lambda: sys.operating_hours * sys.power_utility.rate
        
        def get_GWP_mean_ethanol_price():
            if number in cellulosic_ethanol_configurations:
                f = float(RIN_splitter.split.mean())
                return f * dist.mean_advanced_ethanol_price + (1 - f) * dist.mean_cellulosic_ethanol_price
            else:
                return dist.mean_advanced_ethanol_price
        
        def get_GWP_mean_biodiesel_price():
            if number in cellulosic_configurations:
                f = float(RIN_splitter.split.mean())
                return f * dist.mean_biomass_based_diesel_price + (1 - f) * dist.mean_cellulosic_based_diesel_price
            else:
                return dist.mean_advanced_ethanol_price
        
        sys.define_process_impact(
            key=GWP,
            name='Direct non-biogenic emissions',
            basis='kg',
            inventory=direct_nonbiogenic_emissions,
            CF=1.,
        )
        
        self.flows = {
            'feedstock': feedstock_flow,
            'biodiesel': biodiesel_flow,
            'ethanol': ethanol_flow,
            'natural_gas': natural_gas_flow,
            'crude_glycerol': crude_glycerol_flow,
            'electricity': electricity
        }
        @metric(units='USD/MT')
        def MFPP():
            price = tea.solve_price(feedstock)
            return kg_per_MT * price
        
        @metric(units='USD/L')
        def MESP():
            ethanol_streams = [advanced_ethanol, cellulosic_ethanol]
            ethanol_streams = [i for i in ethanol_streams if not i.isempty()]
            if not ethanol_streams: return 0.
            price = tea.solve_price(ethanol_streams)
            return price / ethanol_L_per_kg
        
        @metric(units='USD/L')
        def MBSP():
            biodiesel_streams = [biomass_based_diesel, cellulosic_based_diesel]
            biodiesel_streams = [i for i in biodiesel_streams if not i.isempty()]
            if not biodiesel_streams: return 0.
            price = tea.solve_price(biodiesel_streams)
            return price / biodiesel_L_per_kg
        
        @metric(units='MT/yr')
        def feedstock_consumption():
            return feedstock_flow()
        
        @metric(units='L/MT')
        def biodiesel_production():
            return biodiesel_flow() / feedstock_consumption.get()
        
        @metric(units='L/ha')
        def biodiesel_yield():
            return biodiesel_production.get() * self.wet_biomass_yield
        
        @metric(units='L/MT')
        def ethanol_production():
            return ethanol_flow() / feedstock_consumption.get()
        
        @metric(units='kWhr/MT')
        def electricity_production():
            value = - electricity() / feedstock_consumption.get()
            if value < 0.: value = 0.
            return value
        
        @metric(units='GGE/MT')
        def net_energy_production():
            GGE_biodiesel_annual = biodiesel_production.get() / 0.9536 / L_per_gal
            GGE_ethanol_annual = ethanol_production.get() / 1.5 / L_per_gal
            GEE_electricity_production = max(-electricity() * 3600 / 114000, 0.) / feedstock_consumption.get()
            GEE_crude_glycerol = crude_glycerol_flow() * 0.1059 / feedstock_consumption.get()
            return (GGE_biodiesel_annual + GGE_ethanol_annual + GEE_electricity_production + GEE_crude_glycerol)
        
        @metric(units='m3/MT')
        def natural_gas_consumption():
            value = natural_gas_flow() / feedstock_consumption.get()
            return value
        
        @metric(units='10^6*USD')
        def TCI():
            return tea.TCI / 1e6 # 10^6*$
        
        @metric(units='%')
        def heat_exchanger_network_error():
            return HXN.energy_balance_percent_error if HXN else 0.
    
        @metric(name='GWP', element='Economic allocation', units='kg*CO2*eq / USD')
        def GWP_economic(): # Cradle to gate
            GWP_material = sys.get_total_feeds_impact(GWP) # kg CO2 eq. / yr
            GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
            sales = (
                biodiesel_flow() * get_GWP_mean_biodiesel_price()
                + ethanol_flow() * get_GWP_mean_ethanol_price()
                + crude_glycerol_flow() * dist.mean_glycerol_price
                + max(-electricity(), 0) * dist.mean_electricity_price
            )
            return (GWP_material + GWP_emissions) / sales
    
        @metric(name='Ethanol GWP', element='Economic allocation', units='kg*CO2*eq / L')
        def GWP_ethanol(): # Cradle to gate
            return GWP_economic.get() * get_GWP_mean_ethanol_price()
        
        @metric(name='Biodiesel GWP', element='Economic allocation', units='kg*CO2*eq / L')
        def GWP_biodiesel(): # Cradle to gate
            if number > 0:
                return GWP_economic.get() * get_GWP_mean_biodiesel_price()
            else:
                return 0.
        
        @metric(name='Crude glycerol GWP', element='Economic allocation', units='kg*CO2*eq / kg')
        def GWP_crude_glycerol(): # Cradle to gate
            if number in biodiesel_configurations:
                return GWP_economic.get() * dist.mean_glycerol_price
            else:
                return 0.
        
        @metric(name='Electricity GWP', element='Economic allocation', units='kg*CO2*eq / MWhr')
        def GWP_electricity(): # Cradle to gate
            if electricity_production.get():
                return GWP_economic.get() * dist.mean_electricity_price * 1000.
    
        @metric(name='Ethanol GWP', element='Displacement allocation', units='kg*CO2*eq / L')
        def GWP_ethanol_displacement(): # Cradle to gate
            if number in ethanol_configurations:
                GWP_material = sys.get_total_feeds_impact(GWP)
                GWP_electricity_production = GWP_characterization_factors['Electricity'] * electricity_production.get() * feedstock_consumption.get()
                GWP_coproducts = sys.get_total_products_impact(GWP)
                GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
                GWP_total = GWP_material + GWP_emissions - GWP_electricity_production - GWP_coproducts # kg CO2 eq. / yr
                return GWP_total / (ethanol_production.get() * feedstock_consumption.get())
            else:
                return 0.
        
        @metric(name='Biodiesel GWP', element='Displacement allocation', units='kg*CO2*eq / L')
        def GWP_biodiesel_displacement(): # Cradle to gate
            if number in biodiesel_configurations:
                GWP_material = sys.get_total_feeds_impact(GWP)
                GWP_electricity_production = GWP_characterization_factors['Electricity'] * electricity_production.get() * feedstock_consumption.get()
                GWP_coproducts = sys.get_total_products_impact(GWP)
                GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
                GWP_total = GWP_material + GWP_emissions - GWP_electricity_production - GWP_coproducts # kg CO2 eq. / yr
                return GWP_total / (biodiesel_production.get() * feedstock_consumption.get())
            else:
                return 0.
        
        # DO NOT DELETE:
        # import thermosteam as tmo
        # glycerol = tmo.Chemical('Glycerol')
        # ethanol = tmo.Chemical('Ethanol')
        # glycerol_GGE = 0.80 * (glycerol.LHV / glycerol.MW) / 121300 # 0.1059 GGE / kg crude-glycerol
        
        @metric(name='Biofuel GWP', element='Energy allocation', units='kg*CO2*eq / GGE')
        def GWP_biofuel_allocation(): # Cradle to gate
            GWP_material = sys.get_total_feeds_impact(GWP)
            GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
            GWP_total = GWP_material + GWP_emissions # kg CO2 eq. / yr
            GGE_biodiesel_annual = (biodiesel_production.get() * feedstock_consumption.get()) / 0.9536 / L_per_gal
            GGE_ethanol_annual = (ethanol_production.get() * feedstock_consumption.get()) / 1.5 / L_per_gal
            GEE_electricity_production = max(-electricity() * 3600 / 114000, 0.) 
            GEE_crude_glycerol = crude_glycerol_flow() * 0.1059
            return GWP_total / (GGE_biodiesel_annual + GGE_ethanol_annual + GEE_electricity_production + GEE_crude_glycerol)
        
        @metric(name='Ethanol GWP', element='Energy allocation', units='kg*CO2*eq / L')
        def GWP_ethanol_allocation(): # Cradle to gate
            return GWP_biofuel_allocation.get() / 1.5 / L_per_gal
        
        @metric(name='Biodiesel GWP', element='Energy allocation', units='kg*CO2*eq / L')
        def GWP_biodiesel_allocation(): # Cradle to gate
            if number > 0:
                return GWP_biofuel_allocation.get() / 0.9536 / L_per_gal
            else:
                return 0.
        
        @metric(name='Crude-glycerol GWP', element='Energy allocation', units='kg*CO2*eq / kg')
        def GWP_crude_glycerol_allocation(): # Cradle to gate
            if number > 0:
                return GWP_biofuel_allocation.get() * 0.1059
            else:
                return 0.
    
        @metric(units='%')
        def IRR():
            return 100. * tea.solve_IRR()
    
        @metric(units='USD/MT')
        def MFPP_derivative():
            if number < 0:
                return 0.
            if self._derivative_disabled:
                return np.nan
            if agile:
                cane_mode.oil_content += 0.01
                sorghum_mode.oil_content += 0.01
            else:
                composition_specification.load_oil_content(composition_specification.oil + 0.01)
            sys.simulate()
            return MFPP.difference()
        
        @metric(units='L/MT')
        def biodiesel_production_derivative():
            if number < 0: return 0.
            if self._derivative_disabled: return np.nan
            return biodiesel_production.difference()
        
        @metric(units='L/MT')
        def ethanol_production_derivative():
            if number < 0: return 0.
            if self._derivative_disabled: return np.nan
            return ethanol_production.difference()
        
        @metric(units='kWhr/MT')
        def electricity_production_derivative():
            if number < 0: return 0.
            if self._derivative_disabled: return np.nan
            return electricity_production.difference()
        
        @metric(units='cf/MT')
        def natural_gas_consumption_derivative():
            if number < 0: return 0.
            if self._derivative_disabled: return np.nan
            return natural_gas_consumption.difference()
        
        @metric(units='10^6*USD')
        def TCI_derivative():
            if number < 0: return 0.
            if self._derivative_disabled: return np.nan
            return TCI.difference()
        
        @metric(name='GWP derivative', element='Economic allocation', units='kg*CO2*eq / USD')
        def GWP_economic_derivative(): # Cradle to gate
            if number < 0: return 0.
            if self._derivative_disabled: return 0.
            return GWP_economic.difference()
    
        @metric(name='GWP derivative', element='Ethanol', units='kg*CO2*eq / L')
        def GWP_ethanol_derivative(): # Cradle to gate
            return GWP_economic_derivative.get() * get_GWP_mean_ethanol_price()
        
        @metric(name='GWP derivative', element='Biodiesel', units='kg*CO2*eq / L')
        def GWP_biodiesel_derivative(): # Cradle to gate
            if number > 0:
                return GWP_economic_derivative.get() * get_GWP_mean_biodiesel_price()
            else:
                return 0.
        
        @metric(name='GWP derivative', element='Crude glycerol', units='kg*CO2*eq / kg')
        def GWP_crude_glycerol_derivative(): # Cradle to gate
            if number > 0:
                return GWP_economic_derivative.get() * dist.mean_glycerol_price
            else:
                return 0.
        
        @metric(name='GWP derivative', element='Electricity', units='kg*CO2*eq / MWhr')
        def GWP_electricity_derivative(): # Cradle to gate
            if electricity_production.get():
                return GWP_economic_derivative.get() * dist.mean_electricity_price * 1000.
            else:
                return 0.
    
        @metric(units='%')
        def ROI():
            return 100. * tea.ROI
        
        # def competitive_biomass_yield_objective(biomass_yield, target):
        #     self.dry_biomass_yield = biomass_yield
        #     F_mass = self.feedstock.F_mass
        #     self.update_feedstock()
        #     specifications = self.sys.specifications
        #     self.sys.specifications = []
        #     try:
        #         self.sys.simulate()
        #     finally:
        #         self.sys.specifications = specifications
        #     self.feedstock.F_mass = F_mass
        #     return 100. * tea.ROI - target
        
        # @metric(name='Competitive biomass yield', element='Feedstock', units='dry MT/ha')
        # def competitive_biomass_yield():
        #     # assert self.ROI_target is None
        #     if self.ROI_target is None: return np.nan
        #     if composition_specification.oil == 0: return self.baseline_dry_biomass_yield
        #     f = competitive_biomass_yield_objective
        #     x0 = 0.1 * self.baseline_dry_biomass_yield
        #     x1 = 3 * self.baseline_dry_biomass_yield
        #     args = (self.ROI_target,)
        #     if (y0:=f(x0, *args)) > 0. or (y1:=f(x1, *args)) < 0.:
        #         return np.nan
        #     else:
        #         return flx.IQ_interpolation(
        #             f, x0, x1, y0, y1, x=competitive_biomass_yield.get(evaluate=False),
        #             args=args, 
        #         )
        
        def NE_and_TCI_at_biomass_yield(dry_biomass_yield):
            self.dry_biomass_yield = dry_biomass_yield
            self.update_feedstock()
            sys.simulate()
            return tea.net_earnings, tea.TCI
        
        def competitive_biomass_yield_objective(biomass_yield, target, mb_NE, An_TCI):
            return 100. * linear_val(biomass_yield, mb_NE) / exponential_val(biomass_yield, An_TCI) - target
        
        @metric(name='Competitive biomass yield', element='Feedstock', units='dry MT/ha')
        def competitive_biomass_yield():
            # assert self.ROI_target is None
            if self.ROI_target is None: return np.nan
            if composition_specification.oil == 0: return self.baseline_dry_biomass_yield
            x0 = self.dry_biomass_yield
            assert x0 < 100, "dry biomass yield over 100 dry MT / ha"
            x1 = 2 * x0
            NE0, TCI0 = tea.net_earnings, tea.TCI
            NE1, TCI1 = NE_and_TCI_at_biomass_yield(x1)
            self.dry_biomass_yield = self.baseline_dry_biomass_yield # Reset dry biomass yield
            mb_NE = linear_fit(x0, x1, NE0, NE1)
            An_TCI = exponential_fit(x0, x1, TCI0, TCI1)
            
            # DO NOT DELETE: This is to validate the functional approximation made here
            # from matplotlib import pyplot as plt
            # N = 10
            # xs = np.linspace(x0, x1, N)
            # NEs, TCIs = zip(*[NE_and_TCI_at_biomass_yield(x) for x in xs])
            # NEs = np.array(NEs)
            # TCIs = np.array(TCIs)
            # plt.figure()
            # plt.plot(xs, NEs, label='NE actual')
            # NEs_model = np.array([linear_val(x, mb_NE) for x in xs])
            # plt.scatter(xs, NEs_model, label='NE model')
            
            # plt.figure()
            # plt.plot(xs, TCIs, label='TCI actual')
            # TCIs_model = np.array([exponential_val(x, An_TCI) for x in xs])
            # plt.scatter(xs, TCIs_model, label='TCI model')
            # dNEs = (NEs - NEs_model)
            
            # TCIs_log = np.log(TCIs)
            # TCIs_model_log = np.log(TCIs_model)
            # dTCIs_log = TCIs_log - TCIs_model_log
            # TCI_log_SSR = (dTCIs_log * dTCIs_log).sum()
            # TCI_ave_log = np.mean(TCIs_log)
            # dTCIs_log_mean = (TCIs_log - TCI_ave_log)
            # TCI_log_SST = (dTCIs_log_mean * dTCIs_log_mean).sum()
            # R2_TCI_log = 1 - TCI_log_SSR / TCI_log_SST
            # print(R2_TCI_log) # 0.9995
            # assert R2_TCI_log > 0.999
            
            # NE_ave = np.mean(NEs)
            # dNEs = NEs - NEs_model
            # dNEs_mean = NEs - NE_ave
            # NE_SSR = (dNEs * dNEs).sum()
            # NE_SST = (dNEs_mean * dNEs_mean).sum()
            # R2_NE = 1 - NE_SSR / NE_SST
            
            # print(R2_NE) # 0.9999
            # assert R2_NE > 0.999
            
            f = competitive_biomass_yield_objective
            args = (self.ROI_target, mb_NE, An_TCI)
            value = flx.IQ_interpolation(
                f, 0.1 * x0, 10 * x1, args=args, 
            )
            return value
        
        @metric(name='Energy competitive biomass yield', element='Feedstock', units='dry MT/ha')
        def energy_competitive_biomass_yield():
            # assert self.net_energy_target is None
            if self.net_energy_target is None: return np.nan
            NEP_target = self.net_energy_target
            if composition_specification.oil == 0: return self.baseline_dry_biomass_yield
            NEP = net_energy_production.cache
            return self.baseline_dry_biomass_yield * NEP / NEP_target  
            
        # def competitive_microbial_oil_yield_objective(microbial_oil_yield, target):
        #     self.update_dry_biomass_yield(self.baseline_dry_biomass_yield)
        #     set_glucose_to_microbial_oil_yield.setter(microbial_oil_yield)
        #     set_xylose_to_microbial_oil_yield.setter(microbial_oil_yield)
        #     self.sys.simulate()
        #     return 100 * tea.ROI - target
        
        # @metric(name='Competitive microbial oil yield', element='Feedstock', units='wt. %')
        # def competitive_microbial_oil_yield():
        #     if self.ROI_target is None or self.microbial_oil_analysis_disactivated: return np.nan
        #     f = competitive_microbial_oil_yield_objective
        #     x0 = 20
        #     x1 = 95
        #     args = (self.ROI_target,)
        #     if (y0:=f(x0, *args)) > 0. or (y1:=f(x1, *args)) < 0.:
        #         return np.nan
        #     else:
        #         flx.IQ_interpolation(
        #             f, x0, x1, y0, y1, xtol=1e-2, ytol=1e-3, args=args,
        #         )
        #         if number in cellulosic_oil_configurations:
        #             return 100. * fermentor.cofermentation[0].product_yield(product='TriOlein', basis='wt')
        #         elif number in oil_configurations:
        #             return 100. * fermentor.fermentation_reaction[0].product_yield(product='TriOlein', basis='wt')
        
        # @metric(name='Energy competitive microbial oil yield', element='Feedstock', units='wt. %')
        # def energy_competitive_microbial_yield():
        #     if self.net_energy_target is None or self.microbial_oil_analysis_disactivated: return np.nan
        #     f = competitive_microbial_oil_yield_objective
        #     x0 = 20
        #     x1 = 95
        #     if (y0:=f(x0)) > 0. or (y1:=f(x1)) < 0.:
        #         return np.nan
        #     else:
        #         flx.IQ_interpolation(
        #             f, x0, x1, y0, y1, xtol=1e-2, ytol=1e-3, args=(self.net_energy_target,)
        #         )
        #         if number in cellulosic_oil_configurations:
        #             return 100. * fermentor.cofermentation[0].product_yield(product='TriOlein', basis='wt')
        #         elif number in oil_configurations:
        #             return 100. * fermentor.fermentation_reaction[0].product_yield(product='TriOlein', basis='wt')
        
        # Single point evaluation for detailed design results
        def set_baseline(p, x=None):
            if x is None:
                x = p.baseline
            else:
                p.baseline = x
            p.setter(x)
        
        if number in cellulosic_ethanol_configurations:
            set_baseline(set_sorghum_glucose_yield, 79)
            set_baseline(set_sorghum_xylose_yield, 86)
            set_baseline(set_cane_glucose_yield, 91.0)
            set_baseline(set_cane_xylose_yield, 97.5)
            set_baseline(set_glucose_to_ethanol_yield, 90)
            set_baseline(set_xylose_to_ethanol_yield, 42)
        if energycane:
            set_baseline(set_cane_oil_content, 2)
        else:
            set_baseline(set_cane_oil_content, self.default_baseline_oil_content)
        set_baseline(set_bagasse_oil_recovery, 70)
        set_baseline(set_juicing_oil_recovery, 60)
        set_baseline(set_crude_oil_price) 
        set_baseline(set_advanced_ethanol_price) 
        set_baseline(set_cellulosic_ethanol_price) 
        set_baseline(set_biomass_based_diesel_price)
        set_baseline(set_cellulosic_based_diesel_price)
        set_baseline(set_crude_glycerol_price, dist.mean_glycerol_price)
        set_baseline(set_natural_gas_price)
        set_baseline(set_electricity_price)
        if number in cellulosic_ethanol_configurations:
            get_stream('ethanol').price = 0.789
        if number > 0:
            set_baseline(set_cane_PL_content, 10)
            set_baseline(set_cane_FFA_content, 10)
        
        for i in model._parameters: setattr(self, i.setter.__name__, i)
        for i in model._metrics: setattr(self, i.getter.__name__, i)
        self.sys = sys
        self.tea = tea
        self.model = model
        self.flowsheet = flowsheet
        self.unit_groups = unit_groups
        self.configuration = configuration
        self.composition_specification = composition_specification
        self.oil_extraction_specification = oil_extraction_specification
        self.available_land = self.baseline_available_land # ha
        self.dry_biomass_yield = self.baseline_dry_biomass_yield # MT / ha
        self.ROI_target = None
        self.net_energy_target = None
        self.microbial_oil_analysis_disactivated = True
        self.__dict__.update(flowsheet.to_dict())
        if feedstock_line: self.set_feedstock_line(feedstock_line)
        if cache is not None: cache[key] = self
        
        # Avoid erros in Monte Carlo of microbial oil production with huge cell
        # mass production
        if number in cellulosic_configurations:
            flowsheet('SludgeCentrifuge').strict_moisture_content = False
            flowsheet(bst.PressureFilter).isplit[
                'Yeast', 'Glucan', 'Xylan', 'Arabinan',
                'Galactan', 'Lignin', 'Solids',
            ] = 1.
        
        ## Simulation
        sys.simulate()
        feedstock.price = tea.solve_price(feedstock)
        if reduce_chemicals: 
            cane_sys.reduce_chemicals()
            cane_sys._load_stream_links()
            sys.simulate()
        
        ## Tests
        try:
            assert len(all_parameter_mockups) == len(model.parameters)
            assert len(all_metric_mockups) == len(model.metrics)
            for mockup, real in zip(all_parameter_mockups, model.parameters):
                assert mockup.index == real.index
            for mockup, real in zip(all_metric_mockups, model.metrics):
                assert mockup.index == real.index
        except:
            breakpoint()
        return self
    
    def oil_recovery(self):
        f = self.flowsheet
        tank = f(bst.BlendingTankWithSkimming)
        recovered_oil = tank.ins[0].F_mass
        microbial_oil = 0
        for reactor in f(bst.StirredTankReactor):
            microbial_oil += sum([i.imass['Oil'] for i in reactor.outs]) - sum([i.imass['Oil'] for i in reactor.ins])
        biomass_oil = self.feedstock.imass['Oil']
        total_oil = microbial_oil + biomass_oil
        return {
            '% Recovery': 100 * recovered_oil / total_oil,
            'Microbial fraction': microbial_oil / total_oil,
            'Biomass fraction': biomass_oil / total_oil,
            'Oil recovery [kg/s]': recovered_oil / 3600,
            'Total oil [kg/s]': total_oil / 3600,
        }
        



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import os
os.environ["NUMBA_DISABLE_JIT"] = '1' # In case numba or numba cache not working properly
import numpy as np
import biosteam as bst
import thermosteam as tmo
import flexsolve as flx
import pytest
from biosteam.process_tools import UnitGroup
from importlib import import_module

__all__ = (
    'test_sugarcane',
    'test_lipidcane',
    'test_cornstover',
    'test_LAOs',
    'test_lactic',
    'test_ethanol_adipic',
    'generate_all_code',
    'generate_code',
    'print_results',
)

feedstocks_by_module = {
    'corn': 'corn',
    'lipidcane': 'lipidcane',
    'cornstover': 'cornstover',
    'sugarcane': 'sugarcane',
    'oilcane': 'feedstock',
    'LAOs': 'glucose',
    'lactic': 'feedstock',
    'ethanol_adipic': 'feedstock',
    'wheatstraw': 'wheatstraw',
    'animal_bedding': 'bedding',
    # 'HP': 'feedstock',
}
products_by_module = {
    'corn': 'ethanol',
    'lipidcane': 'ethanol',
    'cornstover': 'ethanol',
    'sugarcane': 'ethanol',
    'LAOs': 'octene',
    'wheatstraw': 'ethanol',
    'animal_bedding': 'ethanol',
    'lactic': 'lactic_acid',
    'ethanol_adipic': 'ethanol',
    'HP': 'AcrylicAcid',
    'oilcane': 'ethanol',
}

must_load = {
    'oilcane', 'corn', 'sugarcane', 'cornstover', 'lipidcane'
}

marked_slow = {'wheatstraw', 'animal_bedding'}

configurations = {
    'HP': ('cellulosic', 'sugarcane'),
    'oilcane': ('S1', 'S2', 'O1', 'O2', 'O3', 'O4', 'S1*', 'S2*', 'O1*', 'O2*'),
}

def default_settings(f):
    def wrapper(*args, **kwargs):
        bst.process_tools.default()
        try:
            return f(*args, **kwargs)
        finally:
            bst.process_tools.default()
    wrapper.__name__ = f.__name__
    return wrapper
        

def generate_code(module_name, feedstock_name=None, product_name=None, configuration=None):
    if not feedstock_name:
        feedstock_name = feedstocks_by_module[module_name]
    if not product_name:
        product_name = products_by_module[module_name]
    bst.process_tools.default()
    module = import_module('biorefineries.' + module_name)
    if module_name in must_load:
        if configuration is None:
            module.load()
        else:
            module.load(configuration)
    else:
        try: 
            if configuration is None:
                pass
            else:
                module.load(configuration)
        except: pass
    feedstock = getattr(module, feedstock_name)
    product = getattr(module, product_name)
    for tea_name in ('tea', f'{module_name}_tea',  f'{feedstock_name}_tea', f'{product_name}_tea'):
        try:
            tea = getattr(module, tea_name)
        except AttributeError:
            continue
        else:
            break
    try: units = UnitGroup('Biorefinery', tea.units)
    except:
        breakpoint()
    IRR = tea.IRR
    sales = tea.sales
    material_cost = tea.material_cost
    installed_equipment_cost = tea.installed_equipment_cost
    utility_cost = tea.utility_cost
    heating_duty = units.get_heating_duty()
    cooling_duty = units.get_cooling_duty()
    electricity_consumption = units.get_electricity_consumption()
    electricity_production = units.get_electricity_production()
    configuration_tag = f"_{configuration}".replace('*', '_agile') if configuration else ''
    configuration_name = f"'{configuration}'" if configuration else ''
    print(
    ("@pytest.mark.slow\n" if module_name in marked_slow else "") +
    f"""@default_settings
def test_{module_name}{configuration_tag}():
    from biorefineries import {module_name} as module
    if '{module_name}' in must_load:
        module.load({configuration_name})
    else:
        try: module.load({configuration_name})
        except: pass
    feedstock = module.{feedstock_name}
    product = module.{product_name}
    tea = module.{tea_name}
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, {IRR}, rtol=5e-2)
    assert np.allclose(feedstock.price, {feedstock.price}, rtol=5e-2)
    assert np.allclose(product.price, {product.price}, rtol=5e-2)
    assert np.allclose(tea.sales, {sales}, rtol=5e-2)
    assert np.allclose(tea.material_cost, {material_cost}, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, {installed_equipment_cost}, rtol=5e-2)
    assert np.allclose(tea.utility_cost, {utility_cost}, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), {heating_duty}, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), {cooling_duty}, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), {electricity_consumption}, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), {electricity_production}, rtol=5e-2)
    """
    )

def generate_all_code():
    from warnings import filterwarnings
    filterwarnings('ignore')
    for i in feedstocks_by_module: 
        if i in configurations:
            for j in configurations[i]:
                generate_code(i, configuration=j)
        else:
            generate_code(i)
        
def print_results(tea):
    units = UnitGroup('Biorefinery', tea.units)
    print('Sales:', tea.sales)
    print('Material cost:', tea.material_cost)
    print('Installed equipment cost:', tea.installed_equipment_cost)
    print('Utility cost:', tea.utility_cost)
    print('Heating duty:', units.get_heating_duty())
    print('Cooling duty:', units.get_cooling_duty())
    print('Electricity consumption:', units.get_electricity_consumption())
    print('Electricity production:', units.get_electricity_production())

@default_settings
def test_corn():
    from biorefineries import corn as module
    if 'corn' in must_load:
        module.load()
    else:
        try: module.load()
        except: pass
    feedstock = module.corn
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.04922733123672505, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.13227735731092652, rtol=5e-2)
    assert np.allclose(product.price, 0.48547915353569393, rtol=5e-2)
    assert np.allclose(tea.sales, 74995091.73070191, rtol=5e-2)
    assert np.allclose(tea.material_cost, 52916786.73434014, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 65442127.706239425, rtol=5e-2)
    assert np.allclose(tea.utility_cost, 10271412.67910698, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 115.83227253969739, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 105.40843363176639, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 2.034553006809189, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 0.0, rtol=5e-2)
    
@default_settings
def test_lipidcane():
    from biorefineries import lipidcane as module
    if 'lipidcane' in must_load:
        module.load()
    else:
        try: module.load()
        except: pass
    feedstock = module.lipidcane
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.20932919399877384, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 102694047.2272782, rtol=5e-2)
    assert np.allclose(tea.material_cost, 58798737.910276994, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 222740280.64491123, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -34517330.76179105, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 205.17342749647366, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 247.08985731746512, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 7.890958703657048, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 123.19240959446556, rtol=5e-2)
    
@default_settings
def test_cornstover():
    from biorefineries import cornstover as module
    if 'cornstover' in must_load:
        module.load()
    else:
        try: module.load()
        except: pass
    feedstock = module.cornstover
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.05158816935126135, rtol=5e-2)
    assert np.allclose(product.price, 0.7145111617876865, rtol=5e-2)
    assert np.allclose(tea.sales, 131980022.79714778, rtol=5e-2)
    assert np.allclose(tea.material_cost, 82366246.60839348, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 207640203.212064, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -11041041.106571833, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 356.99137288069835, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 373.13690860863034, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 20.4272660859673, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 47.145366085216345, rtol=5e-2)
    
@default_settings
def test_sugarcane():
    from biorefineries import sugarcane as module
    if 'sugarcane' in must_load:
        module.load()
    else:
        try: module.load()
        except: pass
    feedstock = module.sugarcane
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1372113082983206, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 88302441.90067281, rtol=5e-2)
    assert np.allclose(tea.material_cost, 57165746.3360821, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 197624315.82163516, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -17578015.941603705, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 243.72708389992727, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 269.1006983378128, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 8.38858713981407, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 69.47008700698386, rtol=5e-2)
    
@default_settings
def test_oilcane_S1():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('S1')
    else:
        try: module.load('S1')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.038906765965980375, rtol=5e-2)
    assert np.allclose(product.price, 0.7256416231373818, rtol=5e-2)
    assert np.allclose(tea.sales, 81211568.14673989, rtol=5e-2)
    assert np.allclose(tea.material_cost, 64138682.62725732, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 132957617.30555125, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -19877457.122838948, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 301.7835763258438, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 275.28320762945066, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.387905260791424, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 78.81970034013877, rtol=5e-2)
    
@default_settings
def test_oilcane_S2():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('S2')
    else:
        try: module.load('S2')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.038091083249825616, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 141782972.35415757, rtol=5e-2)
    assert np.allclose(tea.material_cost, 74019579.54032566, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 242058951.96357125, rtol=5e-2)
    assert np.allclose(tea.utility_cost, 1856458.8437069105, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 381.73423549500194, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 401.4358937347683, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 24.79328178494203, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 24.793281784942018, rtol=5e-2)
    
@default_settings
def test_oilcane_O1():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O1')
    else:
        try: module.load('O1')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.04381903933083459, rtol=5e-2)
    assert np.allclose(product.price, 0.7256416231373818, rtol=5e-2)
    assert np.allclose(tea.sales, 72541162.23399265, rtol=5e-2)
    assert np.allclose(tea.material_cost, 73561169.3065002, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 167904676.72549188, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -45667863.91251517, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 286.7340545557082, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 270.24956697588044, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.240299491405967, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 162.72194363407567, rtol=5e-2)
    
@default_settings
def test_oilcane_O2():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O2')
    else:
        try: module.load('O2')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.05100049937427724, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 165471446.99165088, rtol=5e-2)
    assert np.allclose(tea.material_cost, 100053205.3827274, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 262103651.71402207, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -5920428.319076779, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 404.00963416799794, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 401.9419701212242, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 23.74726211930081, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 47.52300168041126, rtol=5e-2)
    
@default_settings
def test_oilcane_O3():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O3')
    else:
        try: module.load('O3')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.03870430795144252, rtol=5e-2)
    assert np.allclose(product.price, 0.7256416231373818, rtol=5e-2)
    assert np.allclose(tea.sales, 60154361.83981891, rtol=5e-2)
    assert np.allclose(tea.material_cost, 62950801.382851556, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 160424575.5450812, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -45428439.93708473, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 273.3424982315873, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 255.0361356306883, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 8.942650297243977, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 161.63014291518465, rtol=5e-2)
    
@default_settings
def test_oilcane_O4():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O4')
    else:
        try: module.load('O4')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.04227783727378679, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 147367587.0345435, rtol=5e-2)
    assert np.allclose(tea.material_cost, 82582239.42326018, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 250967289.93424806, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -3658839.0076199747, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 384.3608304399581, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 378.47041113003166, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 23.4408065563754, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 39.86604520866369, rtol=5e-2)
    
@default_settings
def test_oilcane_S1_agile():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('S1*')
    else:
        try: module.load('S1*')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.043169694393018876, rtol=5e-2)
    assert np.allclose(product.price, 0.7256416231373818, rtol=5e-2)
    assert np.allclose(tea.sales, 103452130.03146541, rtol=5e-2)
    assert np.allclose(tea.material_cost, 92199979.56910989, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 139305529.37205353, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -27253989.037229188, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 188.48021056474133, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 209.5042241589957, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.36186124488584, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 92.93539035660477, rtol=5e-2)
    
@default_settings
def test_oilcane_S2_agile():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('S2*')
    else:
        try: module.load('S2*')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.04550810337704206, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 181788090.91717443, rtol=5e-2)
    assert np.allclose(tea.material_cost, 111925631.58243252, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 252483405.54725802, rtol=5e-2)
    assert np.allclose(tea.utility_cost, 1125039.47345165, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 177.93178804843694, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 284.144190858698, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 24.830707961559863, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 36.4696954915187, rtol=5e-2)
    
@default_settings
def test_oilcane_O1_agile():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O1*')
    else:
        try: module.load('O1*')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.04908419340472239, rtol=5e-2)
    assert np.allclose(product.price, 0.7256416231373818, rtol=5e-2)
    assert np.allclose(tea.sales, 93007865.23945123, rtol=5e-2)
    assert np.allclose(tea.material_cost, 106410203.79417554, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 170453639.44925314, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -58827900.327230155, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 241.7199349035914, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 233.27185997786694, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 9.20796433247521, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 148.39641891205272, rtol=5e-2)
    
@default_settings
def test_oilcane_O2_agile():
    from biorefineries import oilcane as module
    if 'oilcane' in must_load:
        module.load('O2*')
    else:
        try: module.load('O2*')
        except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.0589118930242581, rtol=5e-2)
    assert np.allclose(product.price, 0.789, rtol=5e-2)
    assert np.allclose(tea.sales, 210974460.47876817, rtol=5e-2)
    assert np.allclose(tea.material_cost, 146476858.77324435, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 267760597.61769038, rtol=5e-2)
    assert np.allclose(tea.utility_cost, -8520343.481336148, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 233.19182278523107, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 300.35939091803255, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 23.782223104716127, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 54.70678907695694, rtol=5e-2)
    
    
@default_settings
def test_LAOs():
    from biorefineries import LAOs as module
    try: module.load()
    except: pass
    feedstock = module.glucose
    product = module.octene
    tea = module.LAOs_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=5e-2)
    assert np.allclose(feedstock.price, 0.265, rtol=5e-2)
    assert np.allclose(product.price, 1.289861475509996, rtol=5e-2)
    assert np.allclose(tea.sales, 161175673.52938324, rtol=5e-2)
    assert np.allclose(tea.material_cost, 135658829.45874006, rtol=5e-2)
    assert np.allclose(tea.installed_equipment_cost, 54699221.91419212, rtol=5e-2)
    assert np.allclose(tea.utility_cost, 2531383.5854514386, rtol=5e-2)
    assert np.allclose(units.get_heating_duty(), 38.310783441715046, rtol=5e-2)
    assert np.allclose(units.get_cooling_duty(), 123.69592718862404, rtol=5e-2)
    assert np.allclose(units.get_electricity_consumption(), 3.5278713460948774, rtol=5e-2)
    assert np.allclose(units.get_electricity_production(), 3.527871346094877, rtol=5e-2)
    
### DO NOT DELETE:
@pytest.mark.slow
def test_lactic():
    from biorefineries import lactic
    bst.process_tools.default_utilities()
    system = lactic.system
    MPSP = system.lactic_acid.price
    assert np.allclose(MPSP, 1.5285811040727408, atol=0.01)
    tea = system.lactic_tea
    assert np.allclose(tea.sales, 332247972.3322271, rtol=0.01)
    assert np.allclose(tea.material_cost, 218911499.70700422, rtol=0.01)
    assert np.allclose(tea.installed_equipment_cost, 321744540.84435904, rtol=0.01)
    assert np.allclose(tea.utility_cost, 24995492.58140952, rtol=0.01)
    units = UnitGroup('Biorefinery', system.lactic_tea.units)
    assert np.allclose(system.CHP.system_heating_demand/1e6, 1763.4942302374052, rtol=0.01)
    assert np.allclose(-system.CT.system_cooling_water_duty/1e6, 1629.0964343101657, rtol=0.01)
    assert np.allclose(units.get_electricity_consumption(), 45.291535445041546, rtol=0.01)
    assert np.allclose(units.get_electricity_production(), 0.0)

@pytest.mark.slow
def test_ethanol_adipic():
    bst.process_tools.default_utilities()
    from biorefineries import ethanol_adipic
    acid = ethanol_adipic.system_acid
    MESP = acid.ethanol.price * acid._ethanol_kg_2_gal
    assert np.allclose(MESP, 2.486866594017598, atol=0.01)
    tea = acid.ethanol_tea
    assert np.allclose(tea.sales, 152002785.9542236, rtol=0.01)
    assert np.allclose(tea.material_cost, 112973059.93909653, rtol=0.01)
    assert np.allclose(tea.installed_equipment_cost, 202880204.7890376, rtol=0.01)
    assert np.allclose(tea.utility_cost, -18225095.0315181, rtol=0.01)
    units = UnitGroup('Biorefinery', acid.ethanol_tea.units)
    assert np.allclose(acid.CHP.system_heating_demand/1e6, 333.8802418517945, rtol=0.01)
    assert np.allclose(units.get_cooling_duty(), 327.05990128304114, rtol=0.01)
    assert np.allclose(units.get_electricity_consumption(), 23.84999102548279, rtol=0.01)
    assert np.allclose(units.get_electricity_production(), 55.72024685271333, rtol=0.01)
    
    base = ethanol_adipic.system_base
    MESP = base.ethanol.price * base._ethanol_kg_2_gal
    assert np.allclose(MESP, 2.703699730342356, atol=0.01)
    tea = base.ethanol_adipic_tea
    assert np.allclose(tea.sales, 219850046.2308626, rtol=0.01)
    assert np.allclose(tea.material_cost, 120551649.9082640, rtol=0.01)
    assert np.allclose(tea.installed_equipment_cost, 283345546.0556234, rtol=0.01)
    assert np.allclose(tea.utility_cost, 14241964.663629433, rtol=0.01)
    units = UnitGroup('Biorefinery', base.ethanol_adipic_tea.units)
    assert np.allclose(base.CHP.system_heating_demand/1e6, 296.8322623939439, rtol=0.01)
    assert np.allclose(units.get_cooling_duty(), 247.79573940245342, rtol=0.01)
    assert np.allclose(units.get_electricity_consumption(), 26.565278642577344, rtol=0.001)
    assert np.allclose(units.get_electricity_production(), 0.0)

if __name__ == '__main__':
    # generate_all_code()
    test_corn()
    test_sugarcane()
    test_lipidcane()
    test_cornstover()
    test_oilcane_S1()
    test_oilcane_S2()
    test_oilcane_O1()
    test_oilcane_O2()
    test_oilcane_O3()
    test_oilcane_O4()
    test_oilcane_S1_agile()
    test_oilcane_S2_agile()
    test_oilcane_O1_agile()
    test_oilcane_O2_agile()
    # test_LAOs()
    # test_HP_cellulosic()
    # test_HP_sugarcane()
    # test_lactic()
    # test_ethanol_adipic()







# -*- coding: utf-8 -*-
"""
"""
from . import cellulosic_ethanol_tea
from . import conventional_ethanol_tea

__all__ = (
    *cellulosic_ethanol_tea.__all__,
    *conventional_ethanol_tea.__all__,
)

from .cellulosic_ethanol_tea import *
from .conventional_ethanol_tea import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from .. import streams as s
from biorefineries.biodiesel import (
    create_lipid_wash_system,
    create_transesterification_and_biodiesel_separation_system,
)
from biorefineries.cane.systems.juicing import (
    create_feedstock_handling_system, 
    create_juicing_system_up_to_clarification,
)
from biorefineries.cane.systems.sugarcane import (
    create_sucrose_to_ethanol_system,
)

__all__ = (
    'create_juicing_and_lipid_extraction_system',
    'create_lipidcane_to_biodiesel_and_conventional_ethanol_system',
)


# %% Pretreatment section

@SystemFactory(
    ID='juicing_and_lipid_extraction_sys',
    ins=[s.lipidcane,
         *create_juicing_system_up_to_clarification.ins[1:]],
    outs=[s.screened_juice,
          s.lipid,
          s.bagasse,
          s.fiber_fines,
          s.spent_oil_wash_water]
)
def create_juicing_and_lipid_extraction_system(ins, outs, pellet_bagasse=None):
    lipidcane, H3PO4, lime, polymer = ins
    screened_juice, lipid, bagasse, fiber_fines, spent_oil_wash_water = outs
    juicing_sys = create_juicing_system_up_to_clarification(
        pellet_bagasse=pellet_bagasse,
        ins=[lipidcane, H3PO4, lime, polymer], 
        outs=['', bagasse],
        mockup=True,
    )
    clarified_juice, bagasse = juicing_sys.outs
    u = f.unit
    u.U201.isplit['Lipid'] = 0.10 # Crushing mill
    u.S201.isplit['Lipid'] = 1.00 # Fiber screener 
    u.C201.isplit['Lipid'] = 0.98 # Clarifier
    
    # Separate oil and sugar
    T207 = bst.MixTank('T207')
    T207_2 = bst.Splitter('T207_2',
                            split=dict(Lipid=1,
                                       Water=1e-4))
    
    # Screen out small fibers from sugar stream
    S202 = bst.VibratingScreen('S202', outs=(screened_juice, fiber_fines),
                                 split=dict(Ash=1.0,
                                            CaO=1.0,
                                            Cellulose=1.0,
                                            Flocculant=0.0,
                                            Glucose=0.998,
                                            Hemicellulose=1.0,
                                            Lignin=1.0,
                                            Lipid=1.0,
                                            H3PO4=1.0,
                                            Sucrose=0.998,
                                            Water=0.998))
    S202.mesh_opening = 2
    
    lipid_wash_sys = create_lipid_wash_system(
        ins=T207_2-0,
        outs=(lipid, spent_oil_wash_water),
        mockup=True
    )
    
    ### System set-up ###
    clarified_juice-T207-T207_2
    T207-T207_2-1-S202


@SystemFactory(
    ID='lipidcane_sys',
    ins=[*create_juicing_and_lipid_extraction_system.ins,
         create_sucrose_to_ethanol_system.ins[1]],
    outs=[s.ethanol,
          s.biodiesel,
          s.crude_glycerol,
          s.wastewater,
          s.emissions,
          s.ash_disposal]
)
def create_lipidcane_to_biodiesel_and_conventional_ethanol_system(ins, outs):
    
    lipidcane, H3PO4, lime, polymer, denaturant = ins
    ethanol, biodiesel, crude_glycerol, wastewater, emissions, ash_disposal = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        ins=lipidcane,
        mockup=True,
    )
    
    enzyme = bst.Stream(
        ID='enzyme',
        Cellulose=100,
        Water=900,
        units='kg/hr',
        price=0.5
    )
    
    # Hydrolyze lipid bodies
    # Note: It seems odd that there is such a small amount of water in this 
    # treatment, but we stick to the original paper for consistency.
    # Also, it seems odd that the temperature is low during treatment. Regardless,
    # this biorefinery is only used for validation of biosteam.
    T201 = bst.EnzymeTreatment('T201', (feedstock_handling_sys-0, enzyme))
    
    @T201.add_specification(run=True)
    def update_enzyme():
        sugarcane = T201.ins[0]
        F_mass = sugarcane.F_mass
        enzyme.imass['Cellulose', 'Water'] = 0.003 * F_mass * np.array([0.1, 0.9])
    
    ### Oil and juice separation ###
    
    juicing_and_lipid_extraction_sys = create_juicing_and_lipid_extraction_system(
        ins=[T201-0, H3PO4, lime, polymer],
        mockup=True,
    )
    
    ### Ethanol section ###
    
    ethanol_production_sys = create_sucrose_to_ethanol_system(
        ins=[juicing_and_lipid_extraction_sys-0, denaturant],
        outs=[ethanol],
        mockup=True,
    )
    
    ### Biodiesel section ###
    
    # Fresh degummed oil
    oil = juicing_and_lipid_extraction_sys-1
    create_transesterification_and_biodiesel_separation_system(
        ins=oil, 
        outs=[biodiesel, crude_glycerol],
        mockup=True,
    )

    ### Facilities ###
    
    s = f.stream
    u = f.unit
    
    M305 = bst.Mixer('M305', 
        ins=(juicing_and_lipid_extraction_sys-4, 
             juicing_and_lipid_extraction_sys-3,
             *ethanol_production_sys-[1, 2, 3]),
        outs=wastewater,
    )
    
    # Burn bagasse from conveyor belt
    BT = bst.BoilerTurbogenerator('BT',
                                   (juicing_and_lipid_extraction_sys-2, '', 
                                    'boiler_makeup_water', 'natural_gas', '', ''),
                                   (emissions, 'rejected_water_and_blowdown', ash_disposal),
                                   boiler_efficiency=0.80,
                                   turbogenerator_efficiency=0.85)
    
    CT = bst.CoolingTower('CT')
    makeup_water_streams = (s.cooling_tower_makeup_water,
                            s.boiler_makeup_water)
    
    process_water_streams = (s.imbibition_water,
                             s.biodiesel_wash_water,
                             s.rvf_wash_water,
                             s.stripping_water,
                             *makeup_water_streams)
    
    makeup_water = bst.Stream('makeup_water', price=0.000254)
    
    CWP = bst.ChilledWaterPackage('CWP')
    PWC = bst.ProcessWaterCenter('PWC',
                                   (bst.Stream(), makeup_water),
                                   (),
                                   None,
                                   makeup_water_streams,
                                   process_water_streams)
    
    F301 = u.F301
    D303 = u.D303
    HXN = bst.HeatExchangerNetwork('HXN', units=[F301, D303])


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.ethanol import (
    create_beer_distillation_system,
    create_ethanol_purification_system_after_beer_column,
)
from biorefineries.cellulosic import create_facilities
from .biodiesel_ethanol import create_oilcane_to_biodiesel_and_ethanol_1g
from ..fermentation import create_cane_to_combined_1_and_2g_fermentation
from ..lipid_extraction import create_post_fermentation_oil_separation_system
from ... import streams as s

__all__ = (
    'create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation',
)

@SystemFactory(
    ID='oilcane_sys',
    ins=create_oilcane_to_biodiesel_and_ethanol_1g.ins,
    outs=[s.ethanol, s.crude_oil],
)
def create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation(ins, outs, WWT_kwargs=None):
    oilcane, = ins
    ethanol, crude_oil = outs
    
    cane_to_fermentation_sys = create_cane_to_combined_1_and_2g_fermentation('cane_to_fermentation_sys', ins=oilcane)
    beer, lignin, condensate, pretreatment_wastewater, fiber_fines, bagasse_to_boiler = cane_to_fermentation_sys.outs
    cellulosic_beer_distillation_sys = create_beer_distillation_system(
        ins=beer,
        outs=[''],
        mockup=True,
        area=400,
    )
    stripper_process_water = bst.Stream('')
    distilled_beer, stillage = cellulosic_beer_distillation_sys.outs
    ethanol_purification_sys, ep_dct = create_ethanol_purification_system_after_beer_column(
        ins=distilled_beer,
        outs=[ethanol, stripper_process_water],
        mockup=True,
        udct=True,
        area=400,
    )
    post_fermentation_oil_separation_sys, pfls_dct = create_post_fermentation_oil_separation_system(
        ins=stillage, outs=[crude_oil],
        mockup=True,
        area=600,
        udct=True,
        separate_cellmass=True,
    )
    backend_oil, cellmass, wastewater, evaporator_condensate = post_fermentation_oil_separation_sys.outs
    MX_process_water = bst.Mixer(800, (condensate, evaporator_condensate, stripper_process_water),
                                 'recycle_process_water')
    
    bst.Mixer(500,
        ins=[wastewater,
             fiber_fines,
             pretreatment_wastewater,
             evaporator_condensate],
    )
    s = f.stream
    u = f.unit
    MX_solids = bst.Mixer(700, (lignin, cellmass, f.stream.filter_cake, bagasse_to_boiler))
    if WWT_kwargs is None:
        WWT_kwargs = dict(area=500)
    else:
        WWT_kwargs['area'] = 500
    # TODO: Switch this to bst.create_all_facilities
    bst.create_all_facilities(
        feedstock=s.bagasse,
        recycle_process_water_streams=[MX_process_water-0],
        WWT_kwargs=WWT_kwargs,
        CHP_kwargs=dict(ID=700),
        HXN_kwargs=dict(ID=900, ignored=[u.H401, u.H402], Qmin=1e3, acceptable_energy_balance_error=0.01),
        area=800,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>,
#                      Sarang Bhagwat <sarangb2@illinois.edu>,
#                      Yoel Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

import biosteam as bst
from biorefineries.cornstover import CellulosicEthanolTEA

__all__ = ('LacticTEA', 'create_tea',)

class LacticTEA(CellulosicEthanolTEA):
    # For uncertainty analysis
    _TCI_ratio_cached = 1


def create_tea(lactic_sys=None, OSBL_units=None, flowsheet=None):
    flowsheet = flowsheet or bst.main_flowsheet
    sys = flowsheet.system
    u = flowsheet.unit
    s = flowsheet.stream
    lactic_sys = lactic_sys or sys.lactic_sys
    if lactic_sys is not getattr(flowsheet.system, lactic_sys.ID):
        raise RuntimeError(f'The provided system "{lactic_sys.ID}" and '
                            f'flowsheet "{flowsheet.ID}" do not match.')

    if not OSBL_units:
        ISBL_units = set((*sys.pretreatment_sys.units, *sys.conversion_sys.units,
                          *sys.separation_sys.units))
        OSBL_units = list(set(lactic_sys.units).difference(ISBL_units))

    # biosteam Splitters and Mixers have no cost
    for i in OSBL_units:
        if i.__class__ == bst.units.Mixer or i.__class__ == bst.units.Splitter:
            OSBL_units.remove(i)

    # 907.1847 is auom('ton').conversion_factor('kg')
    feedstock = u.U101.ins[0]
    flow_tpd = (feedstock.F_mass-feedstock.imass['H2O'])*24/907.1847*(1-u.U101.divert_ratio)

    lactic_tea = LacticTEA(
            system=lactic_sys, IRR=0.10, duration=(2016, 2046),
            depreciation='MACRS7', income_tax=0.21, operating_days=0.9*365,
            lang_factor=None, construction_schedule=(0.08, 0.60, 0.32),
            startup_months=3, startup_FOCfrac=1, startup_salesfrac=0.5,
            startup_VOCfrac=0.75, WC_over_FCI=0.05,
            finance_interest=0.08, finance_years=10, finance_fraction=0.4,
            OSBL_units=OSBL_units,
            warehouse=0.04, site_development=0.09, additional_piping=0.045,
            proratable_costs=0.10, field_expenses=0.10, construction=0.20,
            contingency=0.10, other_indirect_costs=0.10,
            labor_cost=3212962*flow_tpd/2205,
            labor_burden=0.90, property_insurance=0.007, maintenance=0.03,
            steam_power_depreciation='MACRS20',
            boiler_turbogenerator=u.BT,
            )
    return lactic_tea


# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from biorefineries.cane import create_cellulosic_oilcane_chemicals as create_chemicals

__all__ = ('create_chemicals',)


    



# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst

__all__ = (
    'BlendingTankWithSkimming',
)

# %% Constants

_gal2m3 = 0.003785
_gpm2m3hr = 0.227124
# _m3hr2gpm = 4.40287
_hp2kW = 0.7457
_Gcal2kJ = 4184e3

# %% Biodiesel production

class BlendingTankWithSkimming(bst.MixTank):
    _N_ins = 2
    _N_outs = 2
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 vessel_type=None, tau=None, V_wf=None, 
                 vessel_material=None, kW_per_m3=1.):
        bst.MixTank.__init__(
            self, ID, ins, outs, thermo,
            vessel_type=vessel_type, tau=tau, V_wf=V_wf, 
            vessel_material=vessel_material, kW_per_m3=kW_per_m3
        )
        
    def _run(self):
        feeds = self.ins
        polar_lipids, vegetable_oil = self.outs
        vegetable_oil.mix_from(feeds)
        polar_lipids.copy_thermal_condition(vegetable_oil)
        polar_lipids.imol['PL'] = vegetable_oil.imol['PL'] 
        polar_lipids.imass['Acetone'] = vegetable_oil.imass['PL'] * 2
        vegetable_oil.mol -= polar_lipids.mol




# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.biodiesel import (
    create_lipid_pretreatment_system as create_oil_pretreatment_system,
    create_transesterification_and_biodiesel_separation_system,
)
from .juicing import (
    create_juicing_system,
    create_feedstock_handling_system,
)
from .fermentation import create_sucrose_fermentation_system
from .lipid_extraction import create_lipid_exctraction_system
from .. import units
from .. import streams as s

__all__ = (
    'create_oilcane_to_biodiesel_and_actag_1g',
)

@bst.SystemFactory(
    ID='acTAG_separation_sys',
    ins=[dict(ID='lipid')],
    outs=[dict(ID='acTAG'),
          dict(ID='TAG')],
)
def create_acTAG_separation_system(ins, outs):
    lipid, = ins
    acTAG, TAG = outs
    M1 = bst.Mixer('M1', (lipid, None))
    C1 = units.OleinCrystallizer(
        'C1', M1-0, T=273.15 - 20,
        solid_purity=0.95, 
        melt_purity=0.60,
    )
    PF1 = bst.PressureFilter('PF1', C1-0, [TAG, ''], split=0.5, moisture_content=0.)
    
    def split_phases(unit):
        feed = unit.ins[0]
        solid, liquid = unit.outs
        solid.copy_like(feed['s'])
        liquid.copy_like(feed['l'])
    
    PF1.add_specification(split_phases, args=[PF1])
    
    C2 = units.OleinCrystallizer(
        'C2', PF1-1, T=273.15 - 35,
        solid_purity=0.85, 
        melt_purity=0.95,
    )
    PF2 = bst.PressureFilter('PF2', C2-0, split=0.5, moisture_content=0.)
    PF2.add_specification(split_phases, args=[PF2])
    bst.StorageTank('S2', PF2-1, acTAG, tau=24*7)
    M1.ins[1] = PF2.outs[0]

@SystemFactory(
    ID='oilcane_sys',
    ins=[s.oilcane],
    outs=[s.biodiesel, s.crude_glycerol, s.vinasse, s.acTAG],
)
def create_oilcane_to_biodiesel_and_actag_1g(
            ins, outs,
        ):
    oilcane, = ins
    biodiesel, crude_glycerol, vinasse, acTAG = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        ins=oilcane,
        outs='',
        mockup=True,
        area=100
    )
    
    ### Oil and juice separation ###
    
    juicing_sys, jdct = create_juicing_system(
        ins=feedstock_handling_sys-0,
        outs=['', '', 'fiber_fines'],
        pellet_bagasse=False,
        dry_bagasse=True,
        mockup=True,
        udct=True,
        area=200,
    )
    screened_juice, bagasse, fiber_fines = juicing_sys.outs
    # bagasse_pelleting_sys = create_bagasse_pelleting_system(None, bagasse, area=200, mockup=True)
    # pelleted_bagasse, = bagasse_pelleting_sys.outs
    jdct['S201'].isplit['Lipid'] = 1. # Vibrating screen
    crushing_mill = jdct['U201']
    crushing_mill.tag = "oil extraction"
    crushing_mill.isplit['Lipid'] = 0.90
    fermrxn = tmo.PRxn([
        tmo.Rxn('Glucose -> 2.04 Water + 1.67 CO2 + 0.106 AcetylDiOlein', 'Glucose', 0.156),
        tmo.Rxn('Glucose -> 2.1 Water + 1.72 CO2 + 0.075 TriOlein', 'Glucose', 0.165),
        tmo.Rxn('Glucose -> Cells', 'Glucose', 0.10, basis='wt').copy(basis='mol'),
    ])
    fermentation_sys, epdct = create_sucrose_fermentation_system(
        ins=[screened_juice],
        scrubber=False,
        fermentation_reaction=fermrxn,
        fed_batch=False,
        titer=5.5,
        productivity=0.033,
        product_group='Lipid',
        mockup=True,
        add_urea=True,
        area=300,
        udct=True,
    )
    fermentor = epdct['R301']
    fermentor.N = None
    fermentor.V = 3785.4118
    fermentor.Nmin = 2
    fermentor.Nmax = 36
    product, condensate, vent = fermentation_sys.outs
    
    fermentor.selectivity = 0.75
    fermentor.product_yield = 0.321
    
    @fermentor.add_specification(run=True)
    def update_selectivity_and_product_yield():
        selectivity = fermentor.selectivity
        product_yield = fermentor.product_yield
        fermrxn.X[0] = product_yield * selectivity
        fermrxn.X[1] = product_yield * (1. - selectivity)
    
    post_fermentation_oil_separation_sys, ledct = create_lipid_exctraction_system(
        ins=product,
        outs=[None, None, vinasse],
        mockup=True,
        area=400,
        udct=True
    )
    oil, cellmass, wastewater = post_fermentation_oil_separation_sys.outs
    
    splitter = ledct['U404']
    splitter.lipid_recovery = 0.7
    @splitter.add_specification(run=True)
    def adjust_lipid_recovery():
        total_lipid = fermentor.outs[1].imass['Lipid']
        free_lipid = fermentor.ins[0].imass['Lipid']
        x_free = free_lipid / total_lipid
        splitter.isplit['Lipid'] = 1. - (
            splitter.lipid_recovery * (1 - x_free) + x_free
        )
    
    acTAG_separation_sys, acdct = create_acTAG_separation_system(
        'acTAG_separation_sys', oil, [acTAG, ''], mockup=True, area=500, udct=True
    )
    acTAG, TAG = acTAG_separation_sys.outs
    
    oil_pretreatment_sys, oil_pretreatment_dct = create_oil_pretreatment_system(
        ins=TAG,
        mockup=True,
        outs=['', 'polar_lipids', ''],
        area=600,
        udct=True,
    )
    oil, polar_lipids, wastewater = oil_pretreatment_sys.outs
    
    # Fresh degummed oil
    transesterification_and_biodiesel_separation_sys, tbdct = create_transesterification_and_biodiesel_separation_system(
        ins=oil, 
        outs=[biodiesel, crude_glycerol, ''],
        mockup=True,
        area=600,
        udct=True,
    )
    bst.Mixer(600, [transesterification_and_biodiesel_separation_sys-2, wastewater], 'wastewater')

    ### Facilities ###
    u = f.unit
    bst.create_all_facilities(
        feedstock=bagasse,
        recycle_process_water_streams=(condensate,),
        HXN_kwargs=dict(
            ID=900,
            ignored=lambda: [u.E301, u.D601.boiler, u.D602.boiler, u.H601, 
                             u.H602, u.H603, u.H604, oil_pretreatment_dct['F3'],
                             acdct['C1'], acdct['C2'], ledct['F201']],
            Qmin=1e5,
            acceptable_energy_balance_error=0.01,
        ),
        CHP_kwargs=dict(area=700),
        WWT=False,
        area=800,
    )



# -*- coding: utf-8 -*-
"""
.. autofunction:: biorefineries.cellulosic.systems.create_hot_water_pretreatment_system
.. autofunction:: biorefineries.cellulosic.systems.create_dilute_acid_pretreatment_system
.. autofunction:: biorefineries.cellulosic.systems.create_ammonia_fiber_expansion_pretreatment_system
.. autofunction:: biorefineries.cellulosic.systems.create_alkaline_pretreatment_system

"""
from . import alkaline
from . import ammonia_fiber_expansion
from . import dilute_acid
from . import hot_water

__all__ = (
    *alkaline.__all__,
    *ammonia_fiber_expansion.__all__,
    *dilute_acid.__all__,
    *hot_water.__all__,
)

from .alkaline import *
from .ammonia_fiber_expansion import *
from .dilute_acid import *
from .hot_water import *



# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 14:39:10 2021

@author: yrc2
"""
from . import _feature_mockups as f
from ._parse_configuration import parse_configuration, Configuration, ConfigurationComparison
from warnings import warn
from thermosteam.utils import roundsigfigs
import biorefineries.oilcane as oc
import os
import pandas as pd
import numpy as np
import biosteam as bst
from typing import NamedTuple
from biorefineries.cane.composition import get_composition, set_composition

__all__ = (
    'images_folder',
    'results_folder',
    'feedstocks_folder',
    'get_composition_data',
    'spearman_file',
    'monte_carlo_file',
    'autoload_file_name',
    'get_monte_carlo_across_oil_content',
    'get_monte_carlo',
    'get_line_monte_carlo',
    'load_composition',
    'montecarlo_results',
    'montecarlo_results_short',
    'montecarlo_results_feedstock_comparison',
    'montecarlo_results_configuration_comparison',
    'montecarlo_results_agile_comparison',
    'montecarlo_results_crude_comparison',
    'get_minimum_GWP_reduction',
)

results_folder = os.path.join(os.path.dirname(__file__), 'results')
images_folder = os.path.join(os.path.dirname(__file__), 'images')
feedstocks_folder = os.path.join(os.path.dirname(__file__), 'feedstocks')

# %% Load feedstock data

cane_line_composition_data = None

class OilcaneCompositionPerformance(NamedTuple):
    oil_content: float
    biomass_yield: float

def get_composition_data():
    global cane_line_composition_data
    if cane_line_composition_data is None:
        file = os.path.join(feedstocks_folder, 'cane_composition_data.xlsx')
        data = pd.read_excel(file, header=[0, 1], index_col=[0])
        # Filter poor lines 
        index = data.index
        good_lines = set(index)
        performances = {}
        for name in index:
            line = data.loc[name]
            performance = OilcaneCompositionPerformance(
                line['Stem Oil (dw)']['Mean'],
                line['Dry biomass yield (% WT)']['Mean'],
            )
            performances[name] = performance
        for name, performance in performances.items():
            for other_name, other_performance in performances.items():
                if (name != other_name 
                    and performance.oil_content < other_performance.oil_content
                    and performance.biomass_yield < other_performance.biomass_yield):
                    good_lines.discard(name)
        names = sorted([name for name in index if name in good_lines], 
                       key=lambda name: performances[name].oil_content)
        cane_line_composition_data = data.loc[names]
    return cane_line_composition_data

def load_composition(feedstock, oil, fiber, water, FFA, PL):
    composition = get_composition(feedstock)
    composition['lipid'] = oil
    composition['fiber'] = fiber
    composition['moisture'] = water
    composition['TAG'] = 1. - FFA - PL
    composition['FFA'] = FFA
    composition['PL'] = PL
    del composition['sugar'] # Sugar is backcalculated in `set_composition`
    set_composition(feedstock, **composition)

# %% Load simulation data

def spearman_file(name):
    number, agile = parse_configuration(name)
    filename = f'oilcane_spearman_{number}'
    if agile: filename += '_agile'
    filename += '.xlsx'
    return os.path.join(results_folder, filename)

def monte_carlo_file(name, across_lines=False, extention='xlsx'):
    number, agile, energycane = parse_configuration(name)
    filename = f'oilcane_monte_carlo_{number}'
    if agile: filename += '_agile'
    if energycane: filename += '_energycane'
    if across_lines: filename += '_across_lines'
    filename += '.' + extention
    return os.path.join(results_folder, filename)

def autoload_file_name(name):
    filename = str(name).replace('*', '_agile')
    return os.path.join(results_folder, filename)

def get_monte_carlo_across_oil_content(name, metric, derivative=False):
    key = parse_configuration(name)
    if isinstance(key, Configuration):
        df = pd.read_excel(
            monte_carlo_file(key, True),
            sheet_name=metric if isinstance(metric, str) else metric.short_description,
            index_col=0
        )
    elif isinstance(key, ConfigurationComparison):
        df = (
            get_monte_carlo_across_oil_content(key.a, metric)
            - get_monte_carlo_across_oil_content(key.b, metric)
        )
    else:
        raise Exception('unknown error')
    if derivative: 
        arr = np.diff(df.values) / np.diff(df.columns.values) / 100.
    else:
        arr = df.values
    return arr
        
def get_monte_carlo_key(index, dct, with_units=False):
    key = index[1] if with_units else index[1].split(' [')[0]
    if key in dct: key = f'{key}, {index[0]}'
    return key

def get_line_monte_carlo(line, name, feature, cache={}):
    configuration = parse_configuration(name)
    key = (*configuration, feature.short_description)
    if isinstance(configuration, Configuration):
        if key in cache:
            df = cache[key]
        else:
            file = monte_carlo_file(configuration, across_lines=True)
            cache[key] = df = pd.read_excel(file, header=[0, 1], index_col=[0], sheet_name=feature.short_description)
        mc = df[line]
    elif isinstance(configuration, ConfigurationComparison):
        raise ValueError('name cannot be a configuration comparison')
    else:
        raise Exception('unknown error')
    mc = mc.dropna(how='all', axis=0)
    return mc

def get_monte_carlo(name, features=None, cache={}):
    key = parse_configuration(name)
    if isinstance(key, Configuration):
        if key in cache:
            df = cache[key]
        else:
            file = monte_carlo_file(key)
            cache[key] = df = pd.read_excel(file, header=[0, 1], index_col=[0])
        if features is None:
            mc = df
        elif isinstance(features, bst.Feature):
            mc = df[features.index]
        else:
            mc = df[[i.index for i in features]]
    elif isinstance(key, ConfigurationComparison):
        if features is None:
            features = (
                *f.tea_monte_carlo_metric_mockups, 
                *f.tea_monte_carlo_derivative_metric_mockups,
                *f.lca_monte_carlo_metric_mockups, 
                *f.lca_monte_carlo_derivative_metric_mockups,
                f.GWP_ethanol_displacement,
                f.GWP_ethanol_allocation,
            )
        if isinstance(features, bst.Feature):
            index = features.index
        else:
            index = [i.index for i in features]
        df_a = get_monte_carlo(key.a)[index]
        df_b = get_monte_carlo(key.b)[index]
        row_a = df_a.shape[0]
        row_b = df_b.shape[0]
        try:
            assert row_a == row_b, "shape mismatch"
        except:
            breakpoint()
        mc = df_a - df_b
    else:
        raise Exception('unknown error')
    mc = mc.dropna(how='all', axis=0)
    return mc


def montecarlo_results(with_units=False):
    results = {}    
    for name in oc.configuration_names + oc.comparison_names:
        try: 
            df = get_monte_carlo(name)
        except:
            warn(f'could not load {name}', RuntimeWarning)
            continue
        results[name] = dct = {}
        # if name in ('O1', 'O2'):
        #     index = f.ethanol_over_biodiesel.index
        #     key = get_monte_carlo_key(index, dct, with_units)
        #     data = df[f.ethanol_production.index].values / df[f.biodiesel_production.index].values
        #     q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
        #     dct[key] = {
        #         'mean': np.mean(data),
        #         'std': np.std(data),
        #         'q05': q05,
        #         'q25': q25,
        #         'q50': q50,
        #         'q75': q75,
        #         'q95': q95,
        #     }
        for metric in (*f.tea_monte_carlo_metric_mockups, *f.tea_monte_carlo_derivative_metric_mockups,
                       *f.lca_monte_carlo_metric_mockups, *f.lca_monte_carlo_derivative_metric_mockups,
                       f.GWP_ethanol_displacement, f.GWP_ethanol_allocation):
            index = metric.index
            data = df[index].values
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            key = get_monte_carlo_key(index, dct, with_units)
            dct[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    try:
        df_O2O1 = get_monte_carlo('O2 - O1')
        df_O1 = get_monte_carlo('O1')
    except:
        warn('could not load O2 - O1', RuntimeWarning)
    else:
        results['(O2 - O1) / O1'] = relative_results = {}
        for metric in (f.biodiesel_production, f.ethanol_production):
            index = metric.index
            key = index[1] if with_units else index[1].split(' [')[0]
            data = (df_O2O1[index].values / df_O1[index].values)
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            relative_results[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    return results

# %% Monte carlo values for manuscript

def get_minimum_GWP_reduction():
    L_per_gal = 3.7854
    gal_per_GGE_ethanol = 1.5
    L_per_GGE_ethanol = L_per_gal * gal_per_GGE_ethanol
    gal_per_GGE_biodiesel = 0.9536
    L_per_GGE_biodiesel = L_per_gal * gal_per_GGE_biodiesel
    GWP_ethanol_economic = GWP_biodiesel_economic = 0
    GWP_ethanol_energy = GWP_biodiesel_energy = 0
    GWP_ethanol_economic_index = f.GWP_ethanol.index
    GWP_ethanol_energy_index = f.GWP_ethanol_allocation.index
    GWP_biodiesel_economic_index = f.GWP_biodiesel.index
    GWP_biodiesel_energy_index = f.GWP_biodiesel_allocation.index
    # https://www.epa.gov/sites/default/files/2016-07/documents/select-ghg-results-table-v1.pdf
    GWP_ethanol_dist_and_use = 3.7 * 114000 / 1e6 # kgCO2e per GGE
    GWP_biodiesel_dist_and_use = 3.4 * 114000 / 1e6 # kgCO2e per GGE
    for name in ('O1', 'O2'): 
        df = get_monte_carlo(name)
        GWP_ethanol_economic = max(GWP_ethanol_economic, df[GWP_ethanol_economic_index].max())
        GWP_ethanol_energy = max(GWP_ethanol_energy, df[GWP_ethanol_energy_index].max())
        GWP_biodiesel_economic = max(GWP_biodiesel_economic, df[GWP_biodiesel_economic_index].max())
        GWP_biodiesel_energy = max(GWP_biodiesel_energy, df[GWP_biodiesel_energy_index].max())
    # https://www.epa.gov/sites/default/files/2016-07/documents/select-ghg-results-table-v1.pdf
    GWP_ethanol_economic *= L_per_GGE_ethanol # per L to per GGE
    GWP_ethanol_energy *= L_per_GGE_ethanol # per L to per GGE
    GWP_biodiesel_economic *= L_per_GGE_biodiesel # per L to per GGE
    GWP_biodiesel_energy *= L_per_GGE_biodiesel # per L to per GGE
    GWP_diesel = 11.058 # kgCO2e per GGE
    GWP_gasoline = 11.1948 # kgCO2e per GGE
    return {
        'Ethanol-economic': 100 * (1 - (GWP_ethanol_economic + GWP_ethanol_dist_and_use) / GWP_gasoline),
        'Ethanol-energy': 100 * (1 - (GWP_ethanol_energy + GWP_ethanol_dist_and_use)/ GWP_gasoline),
        'Biodiesel-economic': 100 * (1 - (GWP_biodiesel_economic + GWP_biodiesel_dist_and_use)/ GWP_diesel),
        'Biodiesel-energy': 100 * (1 - (GWP_biodiesel_energy + GWP_biodiesel_dist_and_use) / GWP_diesel),
    }

def montecarlo_results_short(names, metrics=None, derivative=None):
    if metrics is None:
        if derivative:
            metrics = [
                f.MFPP_derivative, f.TCI_derivative, f.ethanol_production_derivative, f.biodiesel_production_derivative, 
                f.electricity_production_derivative, f.natural_gas_consumption_derivative, 
                f.GWP_ethanol_derivative, 
            ]
        else:
            metrics = [
                f.MFPP, f.TCI, f.ethanol_production, f.biodiesel_production, 
                f.electricity_production, f.natural_gas_consumption, f.GWP_ethanol_displacement, 
                f.GWP_ethanol, f.GWP_ethanol_allocation, 
            ]
    results = {}
    for name in names:
        df = get_monte_carlo(name)
        results[name] = dct = {}
        for metric in metrics:
            index = metric.index
            data = df[index].values
            q05, q50, q95 = roundsigfigs(np.percentile(data, [5, 50, 95], axis=0), 3)
            key = get_monte_carlo_key(index, dct, False)
            if q50 < 0:
                dct[key] = f"{-q50} [{-q95}, {-q05}] -negative-"
            else:
                dct[key] = f"{q50} [{q05}, {q95}]"
        if name == 'O2 - O1':
            df_O2O1 = df
            df_O1 = get_monte_carlo('O1')
            for metric in (f.biodiesel_production, f.ethanol_production):
                index = metric.index
                key = f"% {get_monte_carlo_key(index, {}, False)} increase"
                data = 100. * (df_O2O1[index].values / df_O1[index].values)
                q05, q50, q95 = roundsigfigs(np.percentile(data, [5, 50, 95], axis=0), 3)
                if q50 < 0:
                    dct[key] = f"{-q50} [{-q95}, {-q05}] -negative-"
                else:
                    dct[key] = f"{q50} [{q05}, {q95}]"
    return results

def montecarlo_results_feedstock_comparison():
    return montecarlo_results_short(
        names = [
            'O1 - S1',
            'O2 - S2',
        ],
    )
    
def montecarlo_results_configuration_comparison():
    return montecarlo_results_short(
        names = [
            'O2 - O1',
            'O1 - O3',
            'O2 - O4',
        ],
    )

def montecarlo_results_agile_comparison():
    return montecarlo_results_short(
        names = [
            'O1* - O1',
            'O2* - O2',
        ],
    )

def montecarlo_results_crude_comparison():
    return montecarlo_results_short(
        names = [
            'O1 - O3',
            'O2 - O4',
        ],
    )



# -*- coding: utf-8 -*-
"""
.. contents:: :local:
    
.. autoclass:: biorefineries.ethanol.create_beer_distillation_system
.. autoclass:: biorefineries.ethanol.create_ethanol_purification_system_after_beer_column
.. autoclass:: biorefineries.ethanol.create_ethanol_purification_system

References
----------
.. [1] Cortés-Peña et al. BioSTEAM: A Fast and Flexible Platform for the Design,
 Simulation, and Techno-Economic Analysis of Biorefineries under Uncertainty. 
 ACS Sustainable Chem. Eng. 2020. https://doi.org/10.1021/acssuschemeng.9b07040

"""
from .systems import *
from . import systems

__all__ = (
    *systems.__all__,
)
